#!/usr/bin/env python
###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

# the input file format should follow this:
#   path=/Users/emanuele/code/vistrails/branches/v1.2/vistrails/run.py
#   python=/path/to/python
#   env=DYLD_LIBRARY_PATH= 
#   download=http://url.from.where.download.vt.files
#   host=vistrails.sci.utah.edu
#   user=username
#   db=vistrails
#   filename=path_to_vtfile
#   version=598
#   vtid=8
#   port=3306
#   buildalways=true
#   tag=Some text
#   pdf=false
#   workflow=false
#   tree=false
#   getvtl=false
#   embedworkflow=false
#   includefulltree=false
#   other=width=0.45\linewidth 
#
# the buildalways and port lines are optional
# at least a version or a tag must be provided
# if filename is provided host, db, user and vtid are ignored
# please notice that the tag has precedence over version. If a tag is passed
# we will query the database to get the latest version number.

import sys
import os.path
import os
from httplib import HTTP
from urlparse import urlparse
import urllib2
import re
import logging
import shlex
import shutil
import subprocess
import platform

systemType = platform.system()
debug = True
logger = None

###############################################################################

def log(msg):
    global debug, logger
    if debug:
        logger.debug(msg)

###############################################################################
        
def usage():
    print "Usage: "
    print "   %s path/to/options_file" % sys.argv[0]
    sys.exit(1)

###############################################################################
    
def bool_conv(x):
    """bool_conv(x: str) -> bool
    Converts a string to boolean.
    
    """
    s = str(x).upper()
    if s == 'TRUE':
        return True
    if s == 'FALSE':
        return False

###############################################################################

def path_exists_and_not_empty(path):
    """path_exists_and_not_empty(path:str) -> boolean 
    Returns True if given path exists and it's not empty, otherwise returns 
    False.
    
    """
    if os.path.exists(path):
        for root, dirs, file_names in os.walk(path):
            break
        if len(file_names) > 0:
            return True
    return False

###############################################################################

def vtl_exists_in_folder(path):
    """vtl_exists_in_folder(path:str) -> boolean 
    Returns True if given path contains a *.vtl file, otherwise returns False.
    
    """
    for dirpath, dirname,filenames in os.walk(path):
        for f in filenames:
            ext = os.path.splitext(f)[1]
            if ext == ".vtl":
                return True
    return False

###############################################################################

def build_vtl_file(download_url, filename, version, tag, 
                   execute, showspreadsheetonly, embedWorkflow=False,
                   includeFullTree=False):
    log("build_vtl_file")
    try:
        basename = os.path.splitext(os.path.basename(filename))[0]
        if version != None:
            basename += "_%s_%s_%s_%s"%(version,tag,showspreadsheetonly,execute)
        basename += ".vtl"
        vtl_filename = os.path.join(download_url, basename)
        header = '<vtlink filename="%s" version="%s" execute="%s" showSpreadsheetOnly="%s" \
forceDB="%s" tag="%s"/>'%(filename,#os.path.abspath(filename),
              str(version),
              str(execute),
              str(showspreadsheetonly),
              str(False),
              str(tag))
    except Exception, e:
        log("Error: %s"%str(e))
    log("will open %s"%vtl_filename)
    log("vtlfile contents: %s"%header)
    f = open(vtl_filename, "w")
    f.write(header)
    f.close()
    return "run:"+ os.path.abspath(vtl_filename).replace("\\","/")

###############################################################################

def build_vtl_db(download_url, host, db_name, vt_id, version, port, tag, 
                   execute, showspreadsheetonly, embedWorkflow=False, 
                   includeFullTree=False):
    """build_download_vtl_request(host: str, db_name:str, vt_id: str, version: str,
                      port:str, tag: str, execute: bool,
                      showspreadsheetonly: bool)  -> str  """
    log("build_vtl_db")
    try:
        basename = "%s_%s_%s_%s_%s_%s_%s_%s_%s.vtl"%(host,
                                               db_name,
                                               port,
                                               vt_id,
                                               version,
                                               execute,
                                               embedWorkflow,
                                               includeFullTree,
                                               showspreadsheetonly)
        vtl_filename = os.path.join(download_url, basename)
        header = '<vtlink host="%s" database="%s" port="%s" vtid="%s" \
version="%s" execute="%s" showSpreadsheetOnly="%s" forceDB="%s" />'%(
              host,
              db_name,
              str(port),
              str(vt_id),
              str(version),
              str(execute),
              str(showspreadsheetonly),
              str(True))
    except Exception, e:
        log("Error: %s"%str(e))
    log("will open %s"%vtl_filename)
    log("vtlfile contents: %s"%header)
    f = open(vtl_filename, "w")
    f.write(header)
    f.close()
    return "run:"+ os.path.abspath(vtl_filename)

###############################################################################
def download(url,filename, folder=None):
    """download(url:string, filename: string) -> Boolean
    Downloads a binary file from url to filename and returns True (success)
    or False (failure)
    
    """
    try:
        furl = urllib2.urlopen(url)
        if filename is None and folder is not None:
            info = furl.info()
            re_filename = re.compile('attachment;filename=(.*vtl)')
            attach = info.dict['content-disposition']
            filename = os.path.join(folder,
                                    re_filename.match(attach).groups()[0])
            
        log("downloading file: " + filename)
        f = file(filename,'wb')
        f.write(furl.read())
        f.close()
        return True
    except Exception, e:
        log(str(e))
        return False
    
###############################################################################        

def download_as_text(url):
    """download_as_text(url: string) -> string
    Downloads a url as text. It will return None if it failed.
    
    """
    try:
        furl = urllib2.urlopen(url)
        s = furl.read()
        return s
    except:
        return None

###############################################################################

def build_download_vtl_request(download_url, host, db_name, vt_id, version, port, tag, 
                   execute, showspreadsheetonly, embedWorkflow=False, 
                   includeFullTree=False):
    """build_download_vtl_request(host: str, db_name:str, vt_id: str, version: str,
                      port:str, tag: str, execute: bool,
                      showspreadsheetonly: bool)  -> str
        This generates a piece of latex code containing the \href command and
        a \includegraphics command for each image generated.
    """
    
    url_params = "getvt=%s&db=%s&host=%s&port=%s&tag=%s&\
execute=%s&showspreadsheetonly=%s&embedWorkflow=%s&includeFullTree=%s" % (vt_id,
                                                    db_name,
                                                    host,
                                                    port,
                                                    urllib2.quote(tag),
                                                    execute,
                                                    showspreadsheetonly,
                                                    embedWorkflow,
                                                    includeFullTree)
    if version is not None:
        url_params += "&version=%s"%urllib2.quote(version)
    url_params = url_params.replace("%","\%")
    url = "%s?%s"%(download_url, url_params)
    
    return url    

###############################################################################

def _download_content(url, request, path_to_figures):
    """_download_images(url:string, request:string, path_to_figures:string) -> (Boolean, 
                                                                message)
    Downloads all images and pdf files listed in url and saves them to
    path_to_figures
    
    """ 
    website = url
    url = request
    try:
        page = download_as_text(url)
        # we will look for images and other files embedded in the html
        re_imgs = re.compile('<img[^>]*/>')
        re_src = re.compile('(.*src=")([^"]*)"')
        re_a = re.compile('<a[^>]*>[^<]*</a>')
        re_href = re.compile('(.*href=")([^"]*)"')
        images = re_imgs.findall(page)
        files = re_a.findall(page)
        failed = False
        msg = ''
        if len(images) > 0 or len(files) > 0:
            for i in images:
                pngfile = re_src.match(i).groups()[1]
                if not check_url(pngfile):
                    img = website + pngfile
                else:
                    img = pngfile
                    
                if not download(img, os.path.join(path_to_figures,
                                                  os.path.basename(img))):
                    msg += "Error when downloading image: %s. <return>" %img
                    failed = True
                        
            for f in files:
                otherfile = re_href.match(f).groups()[1]
                    
                if not check_url(otherfile):
                    filename = website + otherfile
                else:
                    filename = otherfile
                    
                if not download(filename, os.path.join(path_to_figures,
                                os.path.basename(filename))):
                    msg += "Error when downloading file: %s. <return>"%filename
                    failed = True
                        
            if not failed :
                return (True, "")
            else:
                return (False, msg)
        else:
            msg = "url: '%s' \n returned: %s" % (url,page.strip())
            return (False, msg)
            
    except Exception, e:
        return (False, str(e))
    
def build_vistrails_cmd_line_db(path_to_vistrails, path_to_python,
                                env_for_vistrails, host, 
                                db_name, db_user, vt_id, version, port, 
                                path_to_figures, pdf=False, wgraph=False, 
                                tree=False):
    """ build_vistrails_cmd_line_db(path_to_vistrails: str, env_for_vistrails: str,
                                     host: str, db_name: str, vt_id: str, 
                                     version: str, path_to_figures: str,
                                     pdf:bool, wgraph:bool, tree: bool) -> list
        Build the command line to run vistrails with the given parameters.
    """
    #dump pdf
    if pdf:
        pdfoption = ["-p"]
    else:
        pdfoption = []
    #user
    if db_user is not None and db_user != "":
        useroption = ['-u', '%s'%db_user]
    else:
        useroption = []
    #dump tree and workflow graph
    if wgraph:
        graphoption = ['-G', '%s'%os.path.abspath(path_to_figures)]
    elif tree:
        graphoption = ['-U' ,'%s'%os.path.abspath(path_to_figures)]
    else:
        graphoption = ['-e', '%s'%os.path.abspath(path_to_figures)]
    #don't select a workflow
    if version is not None:
        voption = ":%s"%version
    else:
        voption = ""
        
    prefix = ['%s'%path_to_vistrails]
    if path_to_vistrails.endswith(".py"):
        prefix = ['%s'%path_to_python,
                  '%s'%path_to_vistrails]
    cmd_line = prefix + ['-b'] + graphoption + ['-t', host, '-f',db_name] + \
             useroption + ['-r', port] + pdfoption + [ '%s%s'%(vt_id, voption)]
    return cmd_line

###############################################################################

def build_vistrails_cmd_line_file(path_to_vistrails, path_to_python,
                                  env_for_vistrails, filename, 
                                  version, path_to_figures, pdf=False, 
                                  wgraph=False, tree=False):
    """ build_vistrails_cmd_line_file(path_to_vistrails: str, 
                                      env_for_vostrails: str, filename: str,
                                      version: str, path_to_figures: str,
                                      pdf:bool, wgraph:bool, tree: bool) -> str
        Build the command line to run vistrails with the given parameters.
    """
    #dump pdf
    if pdf:
        pdfoption = ["-p"]
    else:
        pdfoption = []
    #dump tree and workflow graph
    if wgraph:
        graphoption = ['-G','%s'%os.path.abspath(path_to_figures)]
    elif tree:
        graphoption = ['-U','%s'%os.path.abspath(path_to_figures)]
    else:
        graphoption = ['-e','%s'%os.path.abspath(path_to_figures)]
    #don't select a workflow
    if version is not None:
        voption = ":%s"%version
    else:
        voption = ''
    prefix = ['%s'%path_to_vistrails]
    if path_to_vistrails.endswith(".py"):
        prefix = ['%s'%path_to_python,
                  '%s'%path_to_vistrails]
    cmd_line = prefix + ['-b'] + graphoption + pdfoption + \
               [ "%s%s"%(os.path.abspath(filename),voption)]
                      
    return cmd_line

###############################################################################

def generate_latex_db(is_local, download_url, host, db_name, vt_id, version, port, tag, 
                      execute, showspreadsheetonly, path_to_figures, 
                      graphics_options, embedWorkflow=False, 
                      includeFullTree=False):
    """generate_latex(host: str, db_name:str, vt_id: str, version: str,
                      port:str, tag: str, execute: bool,
                      showspreadsheetonly: bool, path_to_figures: str,
                      graphics_options: str, embedWorkflow: bool, 
                      includeFullTree:bool)  -> str
        This generates a piece of latex code containing the \href command and
        a \includegraphics command for each image generated.
    """
    if not is_local:
        if download_url is not None and download_url != "":
            url = build_download_vtl_request(download_url, host, db_name, vt_id, 
                                             version, port, tag, execute, 
                                             showspreadsheetonly, 
                                             embedWorkflow=embedWorkflow,
                                             includeFullTree=includeFullTree)
            href = "\href{%s}{" % url
    else:
        if download_url is not None and download_url != "":
            url = build_vtl_db(download_url, host, db_name, vt_id, 
                            version, port, tag, execute, 
                            showspreadsheetonly, embedWorkflow=embedWorkflow,
                            includeFullTree=includeFullTree)
            href = "\href{%s}{" % url
    ALLOWED_GRAPHICS = [".png", ".jpg", ".pdf"]
    images = []
    for root, dirs, file_names in os.walk(path_to_figures):
        for f in file_names:
            ext = os.path.splitext(f)[1]
            if ext in ALLOWED_GRAPHICS:
                images.append(f)
        break
    n = len(images)
    s = ''
    
    for f in images:
        filename = os.path.join(path_to_figures,f).replace("%","\%")
        if graphics_options:
            s += "\includegraphics[%s]{%s}\n" % (graphics_options, filename.replace("\\","/"))
        else:
            s += "\includegraphics{%s}\n" % filename.replace("\\","/")
    
    if download_url is not None and download_url != "":
        return href + s + "}"
    else:
        return s

###############################################################################

def generate_latex_file(download_url, vtfile, version, tag, 
                      execute, showspreadsheetonly, path_to_figures, 
                      graphics_options,embedWorkflow=False, 
                      includeFullTree=False):
    """generate_latex(host: str, db_name:str, vt_id: str, version: str,
                      port:str, tag: str, execute: bool,
                      showspreadsheetonly: bool, path_to_figures: str,
                      graphics_options: str)  -> str
        This generates a piece of latex code containing the \href command and
        a \includegraphics command for each image generated.
    """
    log("generate_latex_file")
    if download_url is not None and download_url != "":
        url = build_vtl_file(download_url, vtfile, version, tag, execute, 
                             showspreadsheetonly, embedWorkflow=False)
        href = "\href{%s}{" % url
        log(url)
    ALLOWED_GRAPHICS = [".png", ".jpg", ".pdf"]
    images = []
    for root, dirs, file_names in os.walk(path_to_figures):
        for f in file_names:
            ext = os.path.splitext(f)[1]
            if ext in ALLOWED_GRAPHICS:
                images.append(f)
        break
    n = len(images)
    s = ''
    for f in images:
        filename = os.path.join(path_to_figures,f).replace("%","\%")
        if graphics_options:
            s += "\includegraphics[%s]{%s}\n" % (graphics_options, filename.replace("\\","/"))
        else:
            s += "\includegraphics{%s}\n" % filename.replace("\\","/")
    
    if download_url is not None and download_url != "":
        return href + s + "}"
    else:
        return s

###############################################################################

def generate_latex_error(error_msg):
    """ generate_latex_error(error_msg: str) -> str
        this generates a piece of latex code with an error message.
    """
    error_msg = error_msg.replace("\n", "\\MessageBreak ")
    s = """\\PackageError{vistrails}{ An error occurred when executing vistrails. \\MessageBreak
%s
}{}""" % error_msg
    return s

###############################################################################

def run_vistrails_locally_db(path_to_vistrails, path_to_python, 
                             env_for_vistrails, host, db_name, 
                             db_user, vt_id,
                             version, port, path_to_figures, build_always=False,
                             tag='', execute=False, showspreadsheetonly=False,
                             pdf=False, embedWorkflow=False, 
                             includeFullTree=False):
    """run_vistrails_locally_db(path_to_vistrails: str, host: str,
                             db_name: str, vt_id: str, version: str, port: str,
                             path_to_figures: str) -> tuple(bool, str)
        Run vistrails and returns a tuple containing a boolean saying if it was
        successful or not and the latex code.
    """
    v = version
    if tag != '':
        v = tag
    cmd_line = build_vistrails_cmd_line_db(path_to_vistrails, path_to_python,
                                           env_for_vistrails, 
                                           host, db_name, db_user, vt_id,
                                           v, port, path_to_figures, pdf)
    log("run_vistrails_locally_db")
    log("cmdline: %s"%cmd_line)

    if build_always:
        #force to remove the folder so we avoid stale results
        if os.path.exists(path_to_figures):
            shutil.rmtree(path_to_figures)
            
    if not os.path.exists(path_to_figures):
        os.makedirs(path_to_figures)
        
    if build_always or not path_exists_and_not_empty(path_to_figures):
        my_env = os.environ
        sep = ":"
        if systemType in ["Windows", "Microsoft"]:
            sep = ";"
        if env_for_vistrails != '':
            data = shlex.split(env_for_vistrails)
            for d in data:
                (name,val) = d.split("=")
                pos = val.find("$"+name)
                if pos != -1:
                    if pos == 0:
                        my_env[name] = my_env.get(name,'') + \
                                       val[pos+len(name)+1:]
                    else:
                        if pos + len(name) == len(val) -1:
                            my_env[name] = val[0:pos] + my_env.get(name,'')
                        else:
                            my_env[name] = val[0:pos] + my_env.get(name,'') + \
                                           val[pos+len(name)+1:]
                else:
                    my_env[name] = val
        log("env: %s"%my_env)
        if systemType in ['Windows', 'Microsoft']:
            proc = subprocess.Popen(cmd_line,
                                    env=my_env)
        else:
            proc = subprocess.Popen(cmd_line, shell=False,
                                    env=my_env,
                                    stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    close_fds=True)
        proc.wait()
        log("result: %s"%proc.returncode)
        if proc.stdout:
            lines = proc.stdout.readlines()
            log("stdout: %s"%lines)
        if proc.returncode != 0:
            shutil.rmtree(path_to_figures)
            msg = "See vistrails.log for more information."
            return (False, generate_latex_error(msg))
        
    return (True, generate_latex_db(True,os.getcwd(), host, db_name, vt_id, version, port, tag,
                                    execute, showspreadsheetonly, 
                                    path_to_figures, graphics_options, 
                                    embedWorkflow, includeFullTree))

###############################################################################

def run_vistrails_locally_file(path_to_vistrails, path_to_python,
                               env_for_vistrails, filename, 
                               version, path_to_figures, build_always=False,
                               tag='', execute=False, showspreadsheetonly=False,
                               pdf=False, embedWorkflow=False, 
                               includeFullTree=False):
    """run_vistrails_locally_file(path_to_vistrails: str, filename: str,
                                  version: str, path_to_figures: str,
                                  build_always:bool, tag:str, execute:bool,
                                  showspreadsheetonly:bool, pdf:bool) -> 
                                                               tuple(bool, str)
        Run vistrails and returns a tuple containing a boolean saying if it was
        successful or not and the latex code.
    """
    v = version
    if tag != '':
        v = tag
    cmd_line = build_vistrails_cmd_line_file(path_to_vistrails, path_to_python, 
                                             env_for_vistrails,
                                             filename, v, path_to_figures, 
                                             pdf)
    log("run_vistrails_locally_file")
    log("cmdline: %s"%cmd_line)
    if build_always:
        #force to remove the folder so we avoid stale results
        if os.path.exists(path_to_figures):
            shutil.rmtree(path_to_figures)
            
    if not os.path.exists(path_to_figures):
        log("creating image path")
        os.makedirs(path_to_figures)
                
    if build_always or not path_exists_and_not_empty(path_to_figures):
        my_env = os.environ
        sep = ":"
        if systemType in ["Windows", "Microsoft"]:
            sep = ";"
        if env_for_vistrails != '':
            data = shlex.split(env_for_vistrails)
            log("env_for_vistrails: %s"%data)
            for d in data:
                (name,val) = d.split("=")
                pos = val.find("$"+name)
                if pos != -1:
                    if pos == 0:
                        my_env[name] = my_env.get(name,'') + \
                                       val[pos+len(name)+1:]
                    else:
                        if pos + len(name) == len(val) -1:
                            my_env[name] = val[0:pos] + my_env.get(name,'')
                        else:
                            my_env[name] = val[0:pos] + my_env.get(name,'') + \
                                           val[pos+len(name)+1:]
                else:
                    my_env[name] = val
        log("env: %s"%my_env)
        if systemType in ['Windows', 'Microsoft']:
            proc = subprocess.Popen(cmd_line,
                                    env=my_env,
                                    stdout=subprocess.PIPE)
        else:
            proc = subprocess.Popen(cmd_line, shell=False,
                                    env=my_env,
                                    stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    close_fds=True)
        proc.wait()
        log("result: %s"%proc.returncode)
        if proc.stdout:
            lines = proc.stdout.readlines()
            log("stdout: %s"%lines)
        if proc.returncode != 0:
            shutil.rmtree(path_to_figures)
            msg = "See vistrails.log for more information."
            return (False, generate_latex_error(msg))
    log("will generate latex")
    try:
        return (True, generate_latex_file(os.path.dirname(os.path.abspath(filename)), 
                                      filename, version, tag,
                                      execute, showspreadsheetonly,
                                      path_to_figures, graphics_options,
                                      embedWorkflow, includeFullTree))
    except Exception, e:
        log("Error: %s"%str(e))
###############################################################################

def get_vt_graph_locally_db(path_to_vistrails, path_to_python,
                            env_for_vistrails, host, db_name,
                            db_user, vt_id, port, path_to_figures, 
                            build_always=False, pdf=False,
                            embedWorkflow=False, includeFullTree=False):
    """get_vt_graph_locally_db(path_to_vistrails: str, host: str,
                             db_name: str, vt_id: str, port: str,
                             path_to_figures: str, build_always: bool,
                             pdf:bool) -> tuple(bool, str)
        Run vistrails for loading a vistrail and dump the tree to a file and 
        returns a tuple containing a boolean saying if it was
        successful or not and the latex code.
    """
    cmd_line = build_vistrails_cmd_line_db(path_to_vistrails, path_to_python,
                                           env_for_vistrails, 
                                           host, db_name, db_user, vt_id,
                                           None, port, path_to_figures, pdf, False,
                                           True)
    log("get_vt_graph_locally_db")
    log("cmdline: %s"%cmd_line)
    if build_always:
        #force to remove the folder so we avoid stale results
        if os.path.exists(path_to_figures):
            shutil.rmtree(path_to_figures)
            
    if not os.path.exists(path_to_figures):
        os.makedirs(path_to_figures)
    
    if build_always or not path_exists_and_not_empty(path_to_figures):
        my_env = os.environ
        sep = ":"
        if systemType in ["Windows", "Microsoft"]:
            sep = ";"
        if env_for_vistrails != '':
            data = shlex.split(env_for_vistrails)
            for d in data:
                (name,val) = d.split("=")
                pos = val.find("$"+name)
                if pos != -1:
                    if pos == 0:
                        my_env[name] = my_env.get(name,'') + \
                                       val[pos+len(name)+1:]
                    else:
                        if pos + len(name) == len(val) -1:
                            my_env[name] = val[0:pos] + my_env.get(name,'')
                        else:
                            my_env[name] = val[0:pos] + my_env.get(name,'') + \
                                           val[pos+len(name)+1:]
                else:
                    my_env[name] = val
        log("env: %s"%my_env)        
        if systemType in ['Windows', 'Microsoft']:
            proc = subprocess.Popen(cmd_line,
                                    env=my_env,
                                    stdout=subprocess.PIPE)
        else:
            proc = subprocess.Popen(cmd_line, shell=False,
                                    env=my_env,
                                    stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    close_fds=True)
        proc.wait()
        log("result: %s"%proc.returncode)
        if proc.stdout:
            lines = proc.stdout.readlines()
            log("stdout: %s"%lines)
        if proc.returncode != 0:
            os.rmdir(path_to_figures)
            msg = "See vistrails.log for more information."
            return (False, generate_latex_error(msg))

    return (True, generate_latex_db(True, os.getcwd(), host, db_name, vt_id, None, port, '',
                                 False, False,
                                 path_to_figures, graphics_options,
                                 embedWorkflow, includeFullTree))

###############################################################################

def get_vt_graph_locally_file(path_to_vistrails, path_to_python,
                              env_for_vistrails, filename,
                              path_to_figures, build_always=False, pdf=False,
                              embedWorkflow=False, includeFullTree=False):
    """get_vt_graph_locally_file(path_to_vistrails: str, env_for_vistrails: str,
                                 filename: str, version: str, 
                                 path_to_figures: str,build_always: bool,
                                 pdf:bool) -> tuple(bool, str)
        Run VisTrails to load a workflow and dump the graph to a file and 
        returns a tuple containing a boolean saying whether it was
        successful and the latex code.
    """
    cmd_line = build_vistrails_cmd_line_file(path_to_vistrails, 
                                             path_to_python,
                                             env_for_vistrails, 
                                             filename, None, path_to_figures, 
                                             pdf, False, True)
    log("get_vt_graph_locally_file")
    log("cmdline: %s"%cmd_line)
    
    if build_always:
        #force to remove the folder so we avoid stale results
        if os.path.exists(path_to_figures):
            shutil.rmtree(path_to_figures)
            
    if not os.path.exists(path_to_figures):
        log("  creating images folder")
        os.makedirs(path_to_figures)
        
    if build_always or not path_exists_and_not_empty(path_to_figures):
        log("  folder was empty or forcing execution")
        my_env = os.environ
        sep = ":"
        if systemType in ["Windows", "Microsoft"]:
            sep = ";"
        if env_for_vistrails != '':
            data = shlex.split(env_for_vistrails)
            for d in data:
                (name,val) = d.split("=")
                pos = val.find("$"+name)
                if pos != -1:
                    if pos == 0:
                        my_env[name] = my_env.get(name,'') + \
                                       val[pos+len(name)+1:]
                    else:
                        if pos + len(name) == len(val) -1:
                            my_env[name] = val[0:pos] + my_env.get(name,'')
                        else:
                            my_env[name] = val[0:pos] + my_env.get(name,'') + \
                                           val[pos+len(name)+1:]
                else:
                    my_env[name] = val
        log("env: %s"%my_env)
        if systemType in ['Windows', 'Microsoft']:
            proc = subprocess.Popen(cmd_line,
                                    env=my_env,
                                    stdout=subprocess.PIPE)
        else:
            proc = subprocess.Popen(cmd_line, shell=False,
                                    env=my_env,
                                    stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    close_fds=True)
        proc.wait()
        log("result: %s"%proc.returncode)
        if proc.stdout:
            lines = proc.stdout.readlines()
            log("stdout: %s"%lines)
        if proc.returncode != 0:
            os.rmdir(path_to_figures)
            msg = "See vistrails.log and vistrails_run.log for more information."
            return (False, generate_latex_error(msg))
    else:
        log("  found cached images")
        
    return (True, generate_latex_file(os.path.dirname(os.path.abspath(filename)), 
                                      filename, None, '', False, 
                                      False, path_to_figures, 
                                      graphics_options, embedWorkflow, 
                                      includeFullTree))    

###############################################################################    

def get_wf_graph_locally_db(path_to_vistrails, path_to_python,
                            env_for_vistrails, host, db_name,
                            db_user, vt_id, version, port, path_to_figures, 
                            build_always=False, tag='', pdf=False,
                            embedWorkflow=False, includeFullTree=False):
    """get_wf_graph_locally_db(path_to_vistrails: str, host: str,
                             db_name: str, vt_id: str, version: str, port: str,
                             path_to_figures: str,build_always: bool,
                             tag: str, pdf:bool) -> tuple(bool, str)
        Run vistrails for loading a workflow and dump the graph to a file and 
        returns a tuple containing a boolean saying if it was
        successful or not and the latex code.
    """
    v = version
    if tag != '':
        v = tag
    cmd_line = build_vistrails_cmd_line_db(path_to_vistrails, 
                                           path_to_python, env_for_vistrails, 
                                           host, db_name, db_user, vt_id,
                                           v, port, path_to_figures, pdf, 
                                           True, False)
    log("get_wf_graph_locally_db")
    log("cmdline: %s"%cmd_line)
    if build_always:
        #force to remove the folder so we avoid stale results
        if os.path.exists(path_to_figures):
            shutil.rmtree(path_to_figures)
            
    if not os.path.exists(path_to_figures):
        log("  creating images folder")
        os.makedirs(path_to_figures)
        
    if build_always or not path_exists_and_not_empty(path_to_figures):
        log("  folder was empty or forcing execution")
        my_env = os.environ
        sep = ":"
        if systemType in ["Windows", "Microsoft"]:
            sep = ";"
        if env_for_vistrails != '':
            data = shlex.split(env_for_vistrails)
            for d in data:
                (name,val) = d.split("=")
                pos = val.find("$"+name)
                if pos != -1:
                    if pos == 0:
                        my_env[name] = my_env.get(name,'') + \
                                       val[pos+len(name)+1:]
                    else:
                        if pos + len(name) == len(val) -1:
                            my_env[name] = val[0:pos] + my_env.get(name,'')
                        else:
                            my_env[name] = val[0:pos] + my_env.get(name,'') + \
                                           val[pos+len(name)+1:]
                else:
                    my_env[name] = val
        log("env: %s"%my_env)
        if systemType in ['Windows', 'Microsoft']:
            log("env: %s"%my_env)
            proc = subprocess.Popen(cmd_line,
                                    env=my_env,
                                    stdout=subprocess.PIPE)
        else:
            proc = subprocess.Popen(cmd_line, shell=False,
                                    env=my_env,
                                    stdin=subprocess.PIPE,    
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    close_fds=True)
        proc.wait()
        log("result: %s"%proc.returncode)
        if proc.stdout:
            lines = proc.stdout.readlines()
            log("stdout: %s"%lines)
        if proc.returncode != 0:
            os.rmdir(path_to_figures)
            msg = "See vistrails.log for more information."
            return (False, generate_latex_error(msg))
    else:
        log("  found cached images")
        
    return (True, generate_latex_db(True, os.getcwd(), host, db_name, vt_id, version, 
                                    port, tag, execute, showspreadsheetonly,
                                    path_to_figures, graphics_options,
                                    embedWorkflow, includeFullTree))

###############################################################################

def get_wf_graph_locally_file(path_to_vistrails, path_to_python,
                              env_for_vistrails, filename,
                              version, path_to_figures, build_always=False,
                              tag='', pdf=False, embedWorkflow=False, 
                              includeFullTree=False):
    """get_wf_graph_locally_file(path_to_vistrails: str, env_for_vistrails: str,
                                 filename: str, version: str, 
                             path_to_figures: str,build_always: bool,
                             tag: str, pdf:bool) -> tuple(bool, str)
        Run VisTrails to load a workflow and dump the graph to a file and 
        returns a tuple containing a boolean saying whether it was
        successful and the latex code.
    """
    v = version
    if tag != '':
        v = tag
    cmd_line = build_vistrails_cmd_line_file(path_to_vistrails, path_to_python,
                                             env_for_vistrails, 
                                             filename, v, path_to_figures, 
                                             pdf, True, False)
    log("get_wf_graph_locally_file")
    log("cmdline: %s"%cmd_line)
    if build_always:
        #force to remove the folder so we avoid stale results
        if os.path.exists(path_to_figures):
            shutil.rmtree(path_to_figures)
            
    if not os.path.exists(path_to_figures):
        log("  creating images folder")
        os.makedirs(path_to_figures)
        
    if build_always or not path_exists_and_not_empty(path_to_figures):
        log("  folder was empty or forcing execution")
        my_env = os.environ
        sep = ":"
        if systemType in ["Windows", "Microsoft"]:
            sep = ";"
        if env_for_vistrails != '':
            data = shlex.split(env_for_vistrails)
            for d in data:
                (name,val) = d.split("=")
                pos = val.find("$"+name)
                if pos != -1:
                    if pos == 0:
                        my_env[name] = my_env.get(name,'') + \
                                       val[pos+len(name)+1:]
                    else:
                        if pos + len(name) == len(val) -1:
                            my_env[name] = val[0:pos] + my_env.get(name,'')
                        else:
                            my_env[name] = val[0:pos] + my_env.get(name,'') + \
                                           val[pos+len(name)+1:]
                else:
                    my_env[name] = val
        log("env: %s"%my_env)
        if systemType in ['Windows', 'Microsoft']:
            proc = subprocess.Popen(cmd_line,
                                    env=my_env,
                                    stdout=subprocess.PIPE)
        else:
            proc = subprocess.Popen(cmd_line, shell=False,
                                    env=my_env,
                                    stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    close_fds=True)
        proc.wait()
        log("result: %s"%proc.returncode)
        if proc.stdout:
            lines = proc.stdout.readlines()
            log("stdout: %s"%lines)
        if proc.returncode != 0:
            os.rmdir(path_to_figures)
            msg = "See vistrails.log for more information."
            return (False, generate_latex_error(msg))
    else:
        log("  found cached images")
    log("   about to return")    
    return (True, generate_latex_file(os.path.dirname(os.path.abspath(filename)), 
                                      filename, version, tag, execute, 
                                      showspreadsheetonly, path_to_figures, 
                                      graphics_options, embedWorkflow, 
                                      includeFullTree))    

###############################################################################

def run_vistrails_remotely(path_to_vistrails, download_url, host, db_name, vt_id,
                           version, port, path_to_figures, build_always=False,
                           tag='', execute=False, showspreadsheetonly=False,
                           pdf=False, embedWorkflow=False, 
                           includeFullTree=False):
    """run_vistrails_remotely(path_to_vistrails: str, download_url: str host: str,
                              db_name: str, vt_id: str, version: str, port: str,
                              path_to_figures: str, build_always: bool,
                              tag:str, execute: bool, showspreadsheetonly: bool,
                              pdf: bool, embedWorkflow:bool, 
                              includeFullTree:bool)
                                   -> tuple(bool, str)
        Call vistrails remotely to execute a workflow and returns a tuple 
        containing a boolean saying whether it was successful and the latex 
        code.
    """ 
    log("run_vistrails_remotely")
    if build_always:
        #force to remove the folder so we avoid stale results
        if os.path.exists(path_to_figures):
            shutil.rmtree(path_to_figures)
            
    if not path_exists_and_not_empty(path_to_figures) or build_always:
        if not os.path.exists(path_to_figures):
            os.makedirs(path_to_figures)
        
        if check_url(path_to_vistrails):
            website = "://".join(urlparse(path_to_vistrails)[:2])
            request = "?host=%s&db=%s&vt=%s&version=%s&port=%s&pdf=%s" % (host,
                                                                   db_name,
                                                                   vt_id,
                                                                   urllib2.quote(version),
                                                                   port,
                                                                   pdf)
            url = path_to_vistrails + request
            log("will download from: " + url)
            
            (result, msg) = _download_content(website, url, path_to_figures)
            if result == True:
                log("success")
                return (result, generate_latex_db(False, download_url, host, db_name, vt_id, 
                                               version, port, tag, execute,
                                               showspreadsheetonly,
                                               path_to_figures, 
                                               graphics_options,
                                               embedWorkflow, includeFullTree))
            else:
                log("Error: " + msg)
                return (result, generate_latex_error(msg))
        else:
            msg = "Invalid url: %s" % path_to_vistrails
            log("Error: " + msg)
            return (False, generate_latex_error(msg))
    else:
        log("using cached files")
        return (True, generate_latex_db(False, download_url, host, db_name, vt_id, 
                                     version, port, tag, execute, 
                                     showspreadsheetonly,
                                     path_to_figures, graphics_options,
                                     embedWorkflow, includeFullTree))
                
###############################################################################

def get_vt_graph_remotely(path_to_vistrails, download_url, host, db_name, vt_id,
                          port, path_to_figures, build_always=False, pdf=False,
                          embedWorkflow=False, includeFullTree=False):
    """get_vt_graph_remotely(path_to_vistrails: str, download_url: str, host: str,
                              db_name: str, vt_id: str, port: str,
                              path_to_figures: str, build_always: bool, pdf: bool)
                                   -> tuple(bool, str)
        Get the version tree image of a vistrail and returns a tuple containing 
        a boolean saying whether it was successful and the latex code.
    """    
    log("get_vt_graph_remotely")
    if build_always:
        #force to remove the folder so we avoid stale results
        if os.path.exists(path_to_figures):
            shutil.rmtree(path_to_figures)
            
    if not path_exists_and_not_empty(path_to_figures) or build_always:
        if not os.path.exists(path_to_figures):
            os.makedirs(path_to_figures)
        
        if check_url(path_to_vistrails):
            website = "://".join(urlparse(path_to_vistrails)[:2])
            request = "?host=%s&db=%s&vt=%s&port=%s&pdf=%s&showtree=True" % (host,
                                                                   db_name,
                                                                   vt_id,
                                                                   port,
                                                                   pdf)
            url = path_to_vistrails + request
            log("will download from: " + url)
            
            (result, msg) = _download_content(website, url, path_to_figures)
            if result == True:
                log("success")
                return (result, generate_latex_db(False, download_url, host, db_name, 
                                                  vt_id, None, 
                                                  port, '', False,
                                                  False,
                                                  path_to_figures, 
                                                  graphics_options,
                                                  embedWorkflow, includeFullTree))
            else:
                log("Error: " + msg)
                return (result, generate_latex_error(msg))
        else:
            msg = "Invalid url: %s" % path_to_vistrails
            log("Error: " + msg)
            return (False, generate_latex_error(msg))
    else:
        log("using cached files")
        return (True, generate_latex_db(False, download_url, host, db_name, vt_id, 
                                     None, port, '', 
                                     False, False,
                                     path_to_figures, graphics_options,
                                     embedWorkflow, includeFullTree))
        
###############################################################################

def get_wf_graph_remotely(path_to_vistrails, download_url, host, db_name, vt_id,
                           version, port, path_to_figures, build_always=False, 
                           tag='', pdf=False, embedWorkflow=False, 
                           includeFullTree=False):
    """get_wf_graph_remotely(path_to_vistrails: str, download_url:str, host: str,
                              db_name: str, vt_id: str, version: str, port: str,
                              path_to_figures: str, build_always: bool, 
                              tag:str, pdf: bool) -> tuple(bool, str)
        Get the graph of a workflow and returns a tuple containing a boolean 
        saying whether it was successful and the latex code.
    """
    log("get_wf_graph_remotely") 
    
    if build_always:
        #force to remove the folder so we avoid stale results
        if os.path.exists(path_to_figures):
            shutil.rmtree(path_to_figures)
            
    if not path_exists_and_not_empty(path_to_figures) or build_always:
        if not os.path.exists(path_to_figures):
            os.makedirs(path_to_figures)
        
        if check_url(path_to_vistrails):
            website = "://".join(urlparse(path_to_vistrails)[:2])
            request = "?host=%s&db=%s&vt=%s&version=%s&port=%s&pdf=%s&showworkflow=True" % (host,
                                                                   db_name,
                                                                   vt_id,
                                                                   urllib2.quote(version),
                                                                   port,
                                                                   pdf)
            url = path_to_vistrails + request
            log("will download from: " + url)
            
            (result, msg) = _download_content(website, url, path_to_figures)
            if result == True:
                log("success")
                return (result, generate_latex_db(False, download_url, host, db_name, 
                                               vt_id, version, 
                                               port, tag, False,
                                               False,
                                               path_to_figures, 
                                               graphics_options,
                                               embedWorkflow, includeFullTree))
            else:
                log("Error: " + msg)
                return (result, generate_latex_error(msg))
        else:
            msg = "Invalid url: %s" % path_to_vistrails
            log("Error: " + msg)
            return (False, generate_latex_error(msg))
    else:
        log("using cached files")
        return (True, generate_latex_db(False, download_url, host, db_name, vt_id, 
                                     version, port, tag, 
                                     False, False,
                                     path_to_figures, graphics_options,
                                     embedWorkflow, includeFullTree))
            
###############################################################################

def check_path(path):
    """check_path(path:str) -> bool
    Checks if it's a valid system path.
    If path is valid, we will update it to be a realpath.
    
    """
    result = False
    new_path = os.path.realpath(path)
    if os.path.isfile(new_path):
        path = new_path
        result = True
    else:
        result = False
    return result

###############################################################################

def check_url(url):
        try:
            p = urlparse(url)
            h = HTTP(p[1])
            h.putrequest('HEAD', p[2])
            h.endheaders()
            if h.getreply()[0] == 200:
                return True
            else:
                return False
        except Exception, e:
            log(str(e))
            return False

###############################################################################

options_file = None
try:
    options_file = open(sys.argv[1])
except IndexError:
    usage()
    
if debug:
    logger = logging.getLogger("VisTrailsLatex")
    handler = logging.FileHandler('vistrails.log')
    handler.setFormatter(logging.Formatter('VisTrailsLatex - %(asctime)s %(message)s'))
    handler.setLevel(logging.DEBUG)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    log("***********************************")
    log("********* SESSION START ***********")
    
lines = options_file.readlines()
path_to_vistrails = None
path_to_python = None
download_url = None
env_for_vistrails = None
host = None
db_name = None
db_user = None
vt_id = None
filename = None
version = None
graphics_options = None
build_always = False
port = '3306'
version_tag = ''
showspreadsheetonly = False
execute = False
pdf = False
wgraph = False
tree = False
getvtl = False
embedWorkflow=False
includeFullTree=False
for line in lines:
    args = line.split("=")
    if len(args) > 2:
        args[1] = "=".join(args[1:])
    if args[0] == "path":
        path_to_vistrails = args[1].strip(" \n")
    elif args[0] == "python":
        path_to_python = args[1].strip(" \n")
    elif args[0] == "env":
        env_for_vistrails = args[1].strip(" \n")
    elif args[0] == "download":
        download_url = args[1].strip(" \n")
    elif args[0] == "host":
        host = args[1].strip(" \n")
    elif args[0] == "db":
        db_name = args[1].strip(" \n")
    elif args[0] == "user":
        db_user = args[1].strip(" \n")
    elif args[0] == "vtid":
        vt_id = args[1].strip(" \n")
    elif args[0] == "filename":
        filename = args[1].strip(" \n")
    elif args[0] == "version":
        version = args[1].strip(" \n")
    elif args[0] == "port":
        port = args[1].strip(" \n")
    elif args[0] == "buildalways":
        build_always = bool_conv(args[1].strip(" \n"))
    elif args[0] == "tag":
        version_tag = args[1].strip(" \n")
    elif args[0] == "execute":
        execute = bool_conv(args[1].strip(" \n"))
    elif args[0] == "showspreadsheetonly":
        showspreadsheetonly = bool_conv(args[1].strip(" \n"))
    elif args[0] == 'pdf':
        pdf = bool_conv(args[1].strip(" \n"))
    elif args[0] == 'workflow':
        wgraph = bool_conv(args[1].strip(" \n"))
    elif args[0] == 'tree':
        tree = bool_conv(args[1].strip(" \n"))
    elif args[0] == 'getvtl':
        getvtl = bool_conv(args[1].strip(" \n"))
    elif args[0] == 'embedworkflow':
        embedWorkflow = bool_conv(args[1].strip(" \n"))
    elif args[0] == 'includefulltree':
        includeFullTree = bool_conv(args[1].strip(" \n"))
    elif args[0] == "other":
        graphics_options = args[1].strip(" \n")

if showspreadsheetonly:
    execute = True
# then we use the combination host_db_name_port_vt_id_version or
# filename_version to create a unique folder.

# TODO: Maybe we should use a hash of this. For now let's keep it
# legible.
out_type = "png"
if pdf:
    out_type = 'pdf'
if filename is not None:
    path_to_figures = os.path.join("vistrails_images",
                                   "%s_%s_%s" % (filename,
                                                 version,
                                                 out_type))
    if tree:
        #no need to encode version
        path_to_figures = os.path.join("vistrails_images",
                                       "%s_%s" % (filename, out_type))
    elif wgraph:
        path_to_figures = os.path.join("vistrails_images",
                                       "%s_%s_%s_graph" % (filename,
                                                           version,
                                                           out_type)) 
else:
    path_to_figures = os.path.join("vistrails_images",
                               "%s_%s_%s_%s_%s_%s" % (host, db_name, port,
                                                   vt_id,
                                                   version,
                                                   out_type))    
    if tree:
        #no need to encode version
        path_to_figures = os.path.join("vistrails_images",
                                       "%s_%s_%s_%s_%s" % (host, db_name, port,
                                                           vt_id,
                                                           out_type)) 
    elif wgraph:
        path_to_figures = os.path.join("vistrails_images",
                                       "%s_%s_%s_%s_%s_%s_graph" % (host, db_name, port,
                                                                    vt_id,
                                                                    version,
                                                                    out_type))    
# if the path_to_vistrails point to a file that exists on disk, we will use
# it, else let's assume it's a url (we still check if the url is valid inside
# the run_vistrails_remotely function)
log("path_to_figures: %s"%path_to_figures)
if check_path(path_to_vistrails) and filename is None: #run locally
    if tree:
        # we don't need to actually run the workflow, we just get the
        # tree
        log("will run get_vt_graph_locally_db")
        result, latex = get_vt_graph_locally_db(path_to_vistrails,
                                                path_to_python,
                                                env_for_vistrails, host, db_name,
                                                db_user,
                                                vt_id, port, path_to_figures,
                                                build_always, pdf, embedWorkflow,
                                                includeFullTree)
    elif wgraph:
        # we don't need to actually run the workflow, we just get the
        # workflow graph
        log("will run get_wf_graph_locally_db")
        result, latex = get_wf_graph_locally_db(path_to_vistrails,
                                                path_to_python,
                                                env_for_vistrails, host, db_name,
                                                db_user,
                                                vt_id, version, port, path_to_figures,
                                                build_always, version_tag, pdf,
                                                embedWorkflow, includeFullTree)
    else:    
        result, latex = run_vistrails_locally_db(path_to_vistrails,
                                                 path_to_python,
                                                 env_for_vistrails, host, db_name,
                                                 db_user,
                                                 vt_id, version, port, path_to_figures,
                                                 build_always, version_tag, execute,
                                                 showspreadsheetonly, pdf,
                                                 embedWorkflow, includeFullTree)
elif check_path(path_to_vistrails) and filename is not None: #run locally
    if tree:
        # we don't need to actually run the workflow, we just get the
        # tree
        result, latex = get_vt_graph_locally_file(path_to_vistrails,
                                                  path_to_python,
                                                  env_for_vistrails, filename,
                                                  path_to_figures,
                                                  build_always, pdf,
                                                  embedWorkflow, includeFullTree)
    elif wgraph:
        # we don't need to actually run the workflow, we just get the
        # workflow graph
        log("will run get_wf_graph_locally_file")
        result, latex = get_wf_graph_locally_file(path_to_vistrails,
                                                  path_to_python,
                                                  env_for_vistrails, filename,
                                                  version, path_to_figures,
                                                  build_always, version_tag, pdf,
                                                  embedWorkflow, includeFullTree)
    else:    
        result, latex = run_vistrails_locally_file(path_to_vistrails,
                                                   path_to_python,
                                                   env_for_vistrails, filename,
                                                   version, path_to_figures,
                                                   build_always, version_tag, 
                                                   execute, showspreadsheetonly,
                                                   pdf,embedWorkflow, 
                                                   includeFullTree)
elif (not build_always or
      (build_always and check_url(path_to_vistrails))): #run from the web
    if tree:
        result, latex = get_vt_graph_remotely(path_to_vistrails, download_url,
                                              host, db_name,
                                              vt_id, port, path_to_figures,
                                              build_always, pdf, embedWorkflow,
                                              includeFullTree)
    elif wgraph:
        result, latex = get_wf_graph_remotely(path_to_vistrails, download_url,
                                              host, db_name,
                                              vt_id, version, port, path_to_figures,
                                              build_always,version_tag, pdf,
                                              embedWorkflow, includeFullTree)
    else:
        result, latex = run_vistrails_remotely(path_to_vistrails, download_url,
                                               host, db_name,
                                               vt_id, version, port, path_to_figures,
                                               build_always, version_tag, execute,
                                               showspreadsheetonly,pdf,
                                               embedWorkflow, includeFullTree)
    #download vtl file
    if getvtl:
        if (not build_always and not vtl_exists_in_folder(path_to_figures) or
            build_always):
                url = build_download_vtl_request(download_url, host, db_name, 
                                                 vt_id, version, port, version_tag, 
                                                 execute, showspreadsheetonly, 
                                                 embedWorkflow=embedWorkflow, 
                                                 includeFullTree=includeFullTree)
                log("get_vtl %s"%url)
                if download(url, filename=None, folder=path_to_figures) == False:
                    log("Error when downloading vtl to %s"%path_to_figures)
        
                
else:
    result, latex = (False, 
                     generate_latex_error("It is possible that %s is not a valid \
url nor a valid path to vistrails/run.py or that you don't have an internet connection \
and some workflows have the buildalways option. If you already have cached files, \
try removing the buildalways option from vistrails latex command" %\
                     (path_to_vistrails)))
    
# the printed answer will be included inline by the latex compiler.
print latex
if result == True:
    # we will also create a temporary file that can be used to include the
    # images directly
    if not os.path.isdir('cached'):
        if os.path.exists('cached'):
            print generate_latex_error("cached exists and it is not a folder")
            sys.exit(1)
        else:
            os.mkdir('cached')
    if tree:
        texbasename = "%s_%s_%s_%s_%s.tex" % (host, db_name, port, vt_id,out_type)
    elif wgraph:
        texbasename = "%s_%s_%s_%s_%s_%s_graph.tex" % (host, db_name, port, vt_id,
                                                    version, out_type)
    else:
        texbasename = "%s_%s_%s_%s_%s_%s.tex" % (host, db_name, port, vt_id, 
                                              version, out_type)
    texfilename = os.path.join('cached', texbasename)
    texfile = open(texfilename, 'w')
    texfile.write(latex)
    texfile.close()
    log("********** SESSION END ************")
    log("***********************************")
    sys.exit(0)
else:
    log("********** SESSION END ************")
    log("***********************************")
    sys.exit(1)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

import sys
import os
import re
from convert_to_vt import ConvertVTKToVT

def run(in_dir, out_dir):
    def collectFilenames(dir):
        # create the regular expression matcher machines
        fileNameParser = re.compile('.*\.py')

        result = []
        for file in os.listdir(dir):
            # print childDirOrFile
            childDirOrFile = os.path.join(dir,file)
            if os.path.isfile(childDirOrFile):
                # file does match?
                if fileNameParser.match(childDirOrFile):
                    result.append(childDirOrFile)
            elif os.path.isdir(childDirOrFile):
                result += collectFilenames(childDirOrFile)
        return result
    
    all_files = collectFilenames(in_dir)
    in_base = os.path.basename(in_dir)
    for fname in all_files:
        path_before = os.path.dirname(fname)
        path_end = ''
        while os.path.basename(path_before) != in_base:
            if os.path.basename(path_before) != 'Python':
                if path_end == '':
                    path_end = os.path.basename(path_before)
                else:
                    path_end = os.path.join(os.path.basename(path_before), 
                                            path_end)
            path_before = os.path.dirname(path_before)
        out_dirname = os.path.join(out_dir, path_end)

        to_make_dirs = []
        a_dirname = out_dirname
        while not os.path.exists(a_dirname):
            to_make_dirs.append(a_dirname)
            a_dirname = os.path.dirname(a_dirname)
        # print >>sys.stderr, 'to_make_dirs:', to_make_dirs
        for a_dirname in reversed(to_make_dirs):
            # print >>sys.stderr, 'making dir', a_dirname
            os.mkdir(a_dirname)

        script_name = os.path.basename(fname)
        vt_name = os.path.splitext(script_name)[0] + '.vt'
        out_py_name = os.path.join(out_dirname, script_name)
        out_vt_name = os.path.join(out_dirname, vt_name)

        converter = ConvertVTKToVT(fname, out_py_name)
        converter.run()

        print >>sys.stderr, "running 'python %s %s'..." % (out_py_name,
                                                           out_vt_name)
        os.system('python %s %s' % (out_py_name, out_vt_name))

if __name__ == '__main__':
    if len(sys.argv) < 3:
        print 'Usage: %s <in_directory> <out_directory>' % sys.argv[0]
        sys.exit(-1)
    run(*sys.argv[1:])



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.core import query
from vistrails.core.modules.module_registry import get_module_registry
from vistrails.core.utils import append_to_dict_of_lists
import copy
import re

################################################################################

class VisualQuery(query.Query):

    def __init__(self, pipeline, versions_to_check):
        self.queryPipeline = copy.copy(pipeline)
        self.versions_to_check = versions_to_check

    def heuristicDAGIsomorphism(self,
                                target, template,
                                target_ids, template_ids):
        resultIds = set()
        while 1:
            templateNames = set((i, template.modules[i].name)
                                for i in template_ids)
            targetNames = {}
            for i in target_ids:
                append_to_dict_of_lists(targetNames, target.modules[i].name, i)

            nextTargetIds = set()
            nextTemplateIds = set()

            for (i, templateName) in templateNames:
                if templateName not in targetNames:
                    return (False, resultIds)
                else:
                    templateModule = template.modules[i]
                    matched = [tId
                               for tId in targetNames[templateName]
                               if self.matchQueryModule(target.modules[tId],
                                                        templateModule)]
                    resultIds.update(matched)
                    for matchedTargetId in matched:
                        nextTargetIds.update([moduleId for
                                              (moduleId, edgeId) in
                                              target.graph.edges_from(matchedTargetId)])
                    nextTemplateIds.update([moduleId for
                                            (moduleId, edgeId) in
                                            template.graph.edges_from(i)])

            if not len(nextTemplateIds):
                return (True, resultIds)

            target_ids = nextTargetIds
            template_ids = nextTemplateIds

    def run(self, vistrail, name):
        result = []
        self.tupleLength = 2
        for version in self.versions_to_check:
            p = vistrail.getPipeline(version)
            matches = set()
            queryModuleNameIndex = {}
            for moduleId, module in p.modules.iteritems():
                append_to_dict_of_lists(queryModuleNameIndex, module.name, moduleId)
            for querySourceId in self.queryPipeline.graph.sources():
                querySourceName = self.queryPipeline.modules[querySourceId].name
                if not queryModuleNameIndex.has_key(querySourceName):
                    # need to reset matches here!
                    matches = set()
                    continue
                candidates = queryModuleNameIndex[querySourceName]
                atLeastOneMatch = False
                for candidateSourceId in candidates:
                    querySource = self.queryPipeline.modules[querySourceId]
                    candidateSource = p.modules[candidateSourceId]
                    if not self.matchQueryModule(candidateSource,
                                                 querySource):
                        continue
                    (match, targetIds) = self.heuristicDAGIsomorphism \
                                             (template = self.queryPipeline, 
                                              target = p,
                                              template_ids = [querySourceId],
                                              target_ids = [candidateSourceId])
                    if match:
                        atLeastOneMatch = True
                        matches.update(targetIds)
                        
                # We always perform AND operation
                if not atLeastOneMatch:
                    matches = set()
                    break
                
            for m in matches:
                result.append((version, m))
        self.queryResult = result
        self.computeIndices()
        return result
                
    def __call__(self):
        """Returns a copy of itself. This needs to be implemented so that
        a visualquery object looks like a class that can be instantiated
        once per vistrail."""
        return VisualQuery(self.queryPipeline)

    def matchQueryModule(self, template, target):
        """ matchQueryModule(template, target: Module) -> bool        
        Return true if the target module can be matched to the
        template module
        
        """
        if target.name != template.name:
            return False
        if target.getNumFunctions()>template.getNumFunctions():
            return False
        candidateFunctions = {}
        for fid in xrange(template.getNumFunctions()):
            f = template.functions[fid]
            append_to_dict_of_lists(candidateFunctions, f.name, f)

        for f in target.functions:
            if f.name not in candidateFunctions:
                return False
            fNotMatch = True
            candidates = candidateFunctions[f.name]
            for cf in candidates:
                if len(cf.params)!=len(f.params):
                    continue
                pMatch = True
                for pid in xrange(len(cf.params)):
                    cp = cf.params[pid]
                    p = f.params[pid]                    
                    if not self.matchQueryParam(p, cp):
                        pMatch = False
                        break
                if pMatch:
                    fNotMatch = False
                    break
            if fNotMatch:
                return False
        return True

    def matchQueryParam(self, template, target):
        """ matchQueryParam(template: Param, target: Param) -> bool
        Check to see if target can match with a query template
        
        """
        if (template.type != target.type or
            template.identifier != target.identifier or 
            template.namespace != target.namespace):
            return False
        
        reg = get_module_registry()
        desc = reg.get_descriptor_by_name(template.identifier, template.type,
                                          template.namespace)
        return desc.module.query_compute(target.strValue, template.strValue,
                                         template.queryMethod)

        # if template.type=='String':
        #     op = template.queryMethod/2
        #     caseInsensitive = template.queryMethod%2==0
        #     templateStr = template.strValue
        #     targetStr = target.strValue
        #     if caseInsensitive:
        #         templateStr = templateStr.lower()
        #         targetStr = targetStr.lower()

        #     if op==0:
        #         return templateStr in targetStr
        #     if op==1:
        #         return templateStr==targetStr
        #     if op==2:
        #         try:
        #             mo = re.match(templateStr, targetStr)
        #             if mo!=None:
        #                 return mo.end()==len(targetStr)
        #             else:
        #                 return False
        #         except:
        #             return False
        # else:
        #     # FIXME: eval should pretty much never be used
        #     if template.strValue.strip()=='':
        #         return True
        #     realTypeDict = {'Integer': int, 'Float': float}
        #     realType = realTypeDict[template.type]
        #     try:
        #         return realType(template.strValue)==realType(target.strValue)
        #     except: # not a constant
        #         try:
        #             return bool(eval(target.strValue+' '+template.strValue))
        #         except: # not a '<', '>', or '==' expression
        #             try:
        #                 s = template.strValue.replace(' ', '')
        #                 if s[0]=='(':
        #                     mid1 = '<%s)' % target.strValue
        #                 if s[0]=='[':
        #                     mid1 = '<=%s)' % target.strValue
        #                 if s[-1]==')':
        #                     mid2 = '(%s<' % target.strValue
        #                 if s[-1]==']':
        #                     mid2 = '(%s<=' % target.strValue
                            
        #                 s = s.replace(',', '%s and %s' % (mid1, mid2))
        #                 s = '(' + s[1:-1] + ')'
        #                 return eval(s)
        #             except:
        #                 print 'Invalid query "%s".' % template.strValue
        #                 return False



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2013-2014, NYU-Poly.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from file_archive import FileStore
import os

from vistrails.core.system import current_dot_vistrails

from .common import set_default_store, PersistentHash
from .cache import CachedPath, CachedFile, CachedDir
from .queries import QueryCondition, EqualString, EqualInt, IntInRange, \
    Metadata
from .persistedinput import PersistedInputPath, \
    PersistedInputFile, PersistedInputDir
from .queriedinput import QueriedInputPath, QueriedInputFile, QueriedInputDir
from .persistedoutput import PersistedPath, PersistedFile, PersistedDir


def initialize():
    if configuration.check('file_store'):
        file_store_path = configuration.file_store
    else:
        file_store_path = os.path.join(current_dot_vistrails(), 'file_archive')
    if not os.path.exists(file_store_path) or not os.listdir(file_store_path):
        FileStore.create_store(file_store_path)
    set_default_store(FileStore(file_store_path))


_modules = {
        '': [
            # Reference to a specific file
            PersistentHash,

            # Caching modules
            (CachedPath, {'abstract': True}),
            CachedFile,
            CachedDir,

            # Input modules
            (PersistedInputPath, {'abstract': True}),
            PersistedInputFile,
            PersistedInputDir,

            # Query modules
            (QueriedInputPath, {'abstract': True}),
            QueriedInputFile,
            QueriedInputDir,

            # Output modules
            (PersistedPath, {'abstract': True}),
            PersistedFile,
            PersistedDir,
        ],
        'metadata': [
            # Condition & metadata modules
            (QueryCondition, {'abstract': True}),
            (Metadata, {'abstract': True}),
            EqualString,
            EqualInt,
            IntInRange,
        ],
    }


def menu_items():
    from ui import show_viewer
    return [("Show archive content", show_viewer)]



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import copy
from vistrails.db.versions.v0_9_3.domain import DBVistrail, DBAction, DBTag, DBModule, \
    DBConnection, DBPortSpec, DBFunction, DBParameter, DBLocation, DBAdd, \
    DBChange, DBDelete, DBAnnotation, DBPort, DBAbstractionRef, DBGroup, \
    DBWorkflow, DBLog

def translateVistrail(_vistrail):
    def update_key(old_obj, translate_dict):
        return '__notes__'

    def update_annotation(old_obj, translate_dict):
        new_dict = {'DBAnnotation': {'key': update_key}}
        new_list = []
        for annotation in old_obj.db_annotations:
            if annotation.db_key == 'notes':
                new_list.append(DBAnnotation.update_version(annotation, 
                                                            new_dict))
            else:
                new_list.append(DBAnnotation.update_version(annotation,
                                                            {}))
        return new_list
                
    def update_session(old_obj, translate_dict):
        if not old_obj.db_session:
            session = None
        else:
            session = long(old_obj.db_session)
        return session

    def update_workflow(old_obj, translate_dict):
        return DBWorkflow.update_version(old_obj.db_workflow, translate_dict)

    translate_dict = {'DBAction': {'annotations': update_annotation,
                                   'session': update_session},
                      'DBGroup': {'workflow': update_workflow}}
    # pass DBVistrail because domain contains enriched version of the auto_gen
    vistrail = DBVistrail.update_version(_vistrail, translate_dict)
    vistrail.db_version = '0.9.3'
    return vistrail

def translateWorkflow(_workflow):
    def update_workflow(old_obj, translate_dict):
        return DBWorkflow.update_version(old_obj.db_workflow, translate_dict)
    translate_dict = {'DBGroup': {'workflow': update_workflow}}
    workflow = update_workflow(_workflow, translate_dict)
    workflow.db_version = '0.9.3'
    return workflow

def translateLog(_log):
    log = DBLog.update_version(_log, translate_dict)
    log.db_version = '0.9.3'
    return log



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
# Check for testing
""" This module defines the class Module 
"""
from __future__ import division

import copy
from itertools import izip
import weakref

from vistrails.db.domain import DBModule
from vistrails.core.vistrail.annotation import Annotation
from vistrails.core.vistrail.location import Location
from vistrails.core.vistrail.module_control_param import ModuleControlParam
from vistrails.core.vistrail.module_function import ModuleFunction
from vistrails.core.vistrail.module_param import ModuleParam
from vistrails.core.vistrail.port_spec import PortSpec
from vistrails.core.utils import NoSummon
from vistrails.core.modules.module_registry import get_module_registry

import unittest

################################################################################

# A Module stores not only the information, but a method (summon) that
# creates a 'live' object, subclass of core/modules/vistrail_module/Module

class Module(DBModule):
    """ Represents a module from a Pipeline """

    ##########################################################################
    # Constructor and copy

    def __init__(self, *args, **kwargs):
        DBModule.__init__(self, *args, **kwargs)
        if self.cache is None:
            self.cache = 1
        if self.id is None:
            self.id = -1
        if self.location is None:
            self.location = Location(x=-1.0, y=-1.0)
        if self.name is None:
            self.name = ''
        if self.package is None:
            self.package = ''
        if self.version is None:
            self.version = ''
        self.set_defaults()

    def set_defaults(self, other=None):
        if other is None:
            self.portVisible = set()
            self.visible_input_ports = set()
            self.visible_output_ports = set()
            self.connected_input_ports = {}
            self.connected_output_ports = {}
            self.is_valid = False
            self.is_breakpoint = False
            self.is_watched = False
            self._descriptor_info = None
            self._module_descriptor = None
            self.list_depth = 0
            self.iterated_ports = []
        else:
            self.portVisible = copy.copy(other.portVisible)
            self.visible_input_ports = copy.copy(other.visible_input_ports)
            self.visible_output_ports = copy.copy(other.visible_output_ports)
            self.connected_input_ports = copy.copy(other.connected_input_ports)
            self.connected_output_ports = \
                copy.copy(other.connected_output_ports)
            self.is_valid = other.is_valid
            self.is_breakpoint = other.is_breakpoint
            self.is_watched = other.is_watched
            self._descriptor_info = None
            self.list_depth = other.list_depth
            self.iterated_ports = other.iterated_ports
            self._module_descriptor = other._module_descriptor
        if not self.namespace:
            self.namespace = None
        self.function_idx = self.db_functions_id_index
        self.setup_indices()

    def setup_indices(self):
        self._input_port_specs = []
        self._output_port_specs = []
        for port_spec in self.port_spec_list:
            if port_spec.type == 'input':
                self._input_port_specs.append(port_spec)
            elif port_spec.type == 'output':
                self._output_port_specs.append(port_spec)

    def __copy__(self):
        """__copy__() -> Module - Returns a clone of itself"""
        return Module.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBModule.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = Module
        cp.set_defaults(self)
        return cp

    @staticmethod
    def convert(_module):
        if _module.__class__ == Module:
            return
        _module.__class__ = Module
        for _port_spec in _module.db_portSpecs:
            PortSpec.convert(_port_spec)
        if _module.db_location:
            Location.convert(_module.db_location)
        for _function in _module.db_functions:
            ModuleFunction.convert(_function)
        for _annotation in _module.db_get_annotations():
            Annotation.convert(_annotation)
        for _control_parameter in _module.db_get_controlParameters():
            ModuleControlParam.convert(_control_parameter)
        _module.set_defaults()

    ##########################################################################
    # CONSTANTS
        
    VISTRAIL_VAR_ANNOTATION = '__vistrail_var__'
    INLINE_WIDGET_ANNOTATION = '__inline_widgets__'

    ##########################################################################

    id = DBModule.db_id
    cache = DBModule.db_cache
    annotations = DBModule.db_annotations
    control_parameters = DBModule.db_controlParameters
    location = DBModule.db_location
    center = DBModule.db_location
    name = DBModule.db_name
    label = DBModule.db_name
    namespace = DBModule.db_namespace
    package = DBModule.db_package
    version = DBModule.db_version
    port_spec_list = DBModule.db_portSpecs
    internal_version = ''

    # type check this (list, hash)
    def _get_functions(self):
        self.db_functions.sort(key=lambda x: x.db_pos)
        return self.db_functions
    def _set_functions(self, functions):
        # want to convert functions to hash...?
        self.db_functions = functions
    functions = property(_get_functions, _set_functions)
    def add_function(self, function):
        self.db_add_function(function)
    def has_function_with_real_id(self, f_id):
        return self.db_has_function_with_id(f_id)
    def get_function_by_real_id(self, f_id):
        return self.db_get_function_by_id(f_id)

    def add_annotation(self, annotation):
        self.db_add_annotation(annotation)
    def delete_annotation(self, annotation):
        self.db_delete_annotation(annotation)
    def has_annotation_with_key(self, key):
        return self.db_has_annotation_with_key(key)
    def get_annotation_by_key(self, key):
        return self.db_get_annotation_by_key(key)        
    def add_control_parameter(self, controlParameter):
        self.db_add_controlParameter(controlParameter)
    def delete_control_parameter(self, controlParameter):
        self.db_delete_controlParameter(controlParameter)
    def has_control_parameter_with_name(self, name):
        return self.db_has_controlParameter_with_name(name)
    def get_control_parameter_by_name(self, name):
        return self.db_get_controlParameter_by_name(name)
    def toggle_breakpoint(self):
        self.is_breakpoint = not self.is_breakpoint
    def toggle_watched(self):
        self.is_watched = not self.is_watched

    def is_vistrail_var(self):
        return self.has_annotation_with_key(Module.VISTRAIL_VAR_ANNOTATION)
    def get_vistrail_var(self):
        return self.get_annotation_by_key(Module.VISTRAIL_VAR_ANNOTATION).value

    def _get_port_specs(self):
        return self.db_portSpecs_id_index
    port_specs = property(_get_port_specs)
    def has_portSpec_with_name(self, name):
        return self.db_has_portSpec_with_name(name)
    def get_portSpec_by_name(self, name):
        return self.db_get_portSpec_by_name(name)
    def add_port_spec(self, spec):
        DBModule.db_add_portSpec(self, spec)
        if spec.type == 'input':
            self._input_port_specs.append(spec)
        elif spec.type == 'output':
            self._output_port_specs.append(spec)
    # override DBModule.db_add_portSpec so that _*_port_specs are updated
    db_add_portSpec = add_port_spec
    def delete_port_spec(self, spec):
        if spec.type == 'input':
            self._input_port_specs.remove(spec)
        elif spec.type == 'output':
            self._output_port_specs.remove(spec)
        DBModule.db_delete_portSpec(self, spec)
    # override DBModule.db_delete_portSpec so that _*_port_specs are updated
    db_delete_portSpec = delete_port_spec

    def _get_input_port_specs(self):
        return sorted(self._input_port_specs, 
                      key=lambda x: (x.sort_key, x.id))
    input_port_specs = property(_get_input_port_specs)
    def _get_output_port_specs(self):
        return sorted(self._output_port_specs, 
                      key=lambda x: (x.sort_key, x.id))
    output_port_specs = property(_get_output_port_specs)

    def _get_descriptor_info(self):
        if self._descriptor_info is None:
            self._descriptor_info = (self.package, self.name, 
                                     self.namespace, self.version,
                                     str(self.internal_version))
        return self._descriptor_info
    descriptor_info = property(_get_descriptor_info)

    def _get_module_descriptor(self):
        if self._module_descriptor is None or \
                self._module_descriptor() is None:
            reg = get_module_registry()
            self._module_descriptor = \
                weakref.ref(reg.get_descriptor_by_name(*self.descriptor_info))
        return self._module_descriptor()
    def _set_module_descriptor(self, descriptor):
        self._module_descriptor = weakref.ref(descriptor)
    module_descriptor = property(_get_module_descriptor, 
                                 _set_module_descriptor)

    def _get_editable_input_ports(self):
        if self.has_annotation_with_key(Module.INLINE_WIDGET_ANNOTATION):
            values = self.get_annotation_by_key(
                             Module.INLINE_WIDGET_ANNOTATION).value.split(',')
            return set() if values == [''] else set(values)
        elif get_module_registry(
                   ).is_constant_module(self.module_descriptor.module):
            # Show value by default on constant modules
            return set(['value'])
        return set()
    editable_input_ports = property(_get_editable_input_ports)

    def get_port_spec(self, port_name, port_type):
        """get_port_spec(port_name: str, port_type: str: ['input' | 'output'])
             -> PortSpec

        """
        if self.has_portSpec_with_name((port_name, port_type)):
            return self.get_portSpec_by_name((port_name, port_type))
        desc = self.module_descriptor
        reg = get_module_registry()
        return reg.get_port_spec_from_descriptor(desc, port_name, port_type)

    def has_port_spec(self, port_name, port_type):
        if self.has_portSpec_with_name((port_name, port_type)):
            return True
        reg = get_module_registry()
        desc = self.module_descriptor
        return reg.has_port_spec_from_descriptor(desc, port_name, port_type)

    def summon(self):
        result = self.module_descriptor.module()
        result.transfer_attrs(self)

        # FIXME this may not be quite right because we don't have self.registry
        # anymore.  That said, I'm not sure how self.registry would have
        # worked for hybrids...
        result.registry = get_module_registry()
        return result

    def is_group(self):
        return False
    def is_abstraction(self):
        return False

    def getNumFunctions(self):
        """getNumFunctions() -> int - Returns the number of functions """
        return len(self.functions)

    def sourcePorts(self):
        """sourcePorts() -> list of Port 
        Returns list of source (output) ports module supports.

        """
        registry = get_module_registry()
        desc = self.module_descriptor
        ports = registry.module_source_ports_from_descriptor(True, desc)
        ports.extend(self.output_port_specs)
        return ports
    
    def destinationPorts(self):
        """destinationPorts() -> list of Port 
        Returns list of destination (input) ports module supports

        """
        registry = get_module_registry()
        desc = self.module_descriptor
        ports = registry.module_destination_ports_from_descriptor(True, desc)
        ports.extend(self.input_port_specs)
        return ports

    ##########################################################################
    # Debugging

    def show_comparison(self, other):
        if type(other) != type(self):
            print "Type mismatch"
            print type(self), type(other)
        elif self.id != other.id:
            print "id mismatch"
            print self.id, other.id
        elif self.name != other.name:
            print "name mismatch"
            print self.name, other.name
        elif self.cache != other.cache:
            print "cache mismatch"
            print self.cache, other.cache
        elif self.location != other.location:
            print "location mismatch"
            # FIXME Location has no show_comparison
            # self.location.show_comparison(other.location)
        elif len(self.functions) != len(other.functions):
            print "function length mismatch"
            print len(self.functions), len(other.functions)
        else:
            for f, g in izip(self.functions, other.functions):
                if f != g:
                    print "function mismatch"
                    f.show_comparison(g)
                    return
            print "No difference found"
            assert self == other

    ##########################################################################
    # Operators

    def __str__(self):
        """__str__() -> str Returns a string representation of itself. """
        def get_name():
            if self.namespace:
                return self.namespace + '|' + self.name
            return self.name
        return ("(Module '%s:%s' id=%s functions:%s port_specs:%s annotations:%s control_parameters:%s)@%X" %
                (self.package,
                 get_name(),
                 self.id,
                 [str(f) for f in self.functions],
                 [str(port_spec) for port_spec in self.db_portSpecs],
                 [str(a) for a in self.annotations],
                 [str(c) for c in self.control_parameters],
                 id(self)))

    def __eq__(self, other):
        """ __eq__(other: Module) -> boolean
        Returns True if self and other have the same attributes. Used by == 
        operator. 
        
        """
        if type(other) != type(self):
            return False
        if self.name != other.name:
            return False
        if self.namespace != other.namespace:
            return False
        if self.package != other.package:
            return False
        if self.cache != other.cache:
            return False
        if self.location != other.location:
            return False
        if len(self.functions) != len(other.functions):
            return False
        if len(self.annotations) != len(other.annotations):
            return False
        if len(self.control_parameters) != len(other.control_parameters):
            return False
        for f, g in izip(self.functions, other.functions):
            if f != g:
                return False
        for f, g in izip(self.annotations, other.annotations):
            if f != g:
                return False
        for f, g in izip(self.control_parameters, other.control_parameters):
            if f != g:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    ##########################################################################
    # Properties


################################################################################
# Testing


class TestModule(unittest.TestCase):

    def create_module(self, id_scope=None):
        from vistrails.core.modules.basic_modules import identifier as basic_pkg
        from vistrails.db.domain import IdScope
        if id_scope is None:
            id_scope = IdScope()
        
        params = [ModuleParam(id=id_scope.getNewId(ModuleParam.vtType),
                                  type='Int',
                                  val='1')]
        functions = [ModuleFunction(id=id_scope.getNewId(ModuleFunction.vtType),
                                    name='value',
                                    parameters=params)]
        control_parameters = [ModuleControlParam(id=id_scope.getNewId(ModuleControlParam.vtType),
                                  name='combiner',
                                  value='pairwise')]
        module = Module(id=id_scope.getNewId(Module.vtType),
                        name='Float',
                        package=basic_pkg,
                        functions=functions,
                        controlParameters=control_parameters)
        return module

    def test_copy(self):
        """Check that copy works correctly"""
        from vistrails.db.domain import IdScope
        
        id_scope = IdScope()
        m1 = self.create_module(id_scope)
        m2 = copy.copy(m1)
        self.assertEquals(m1, m2)
        self.assertEquals(m1.id, m2.id)
        m3 = m1.do_copy(True, id_scope, {})
        self.assertEquals(m1, m3)
        self.assertNotEquals(m1.id, m3.id)

    def test_serialization(self):
        """ Check that serialize and unserialize are working properly """
        import vistrails.core.db.io

        m1 = self.create_module()
        xml_str = vistrails.core.db.io.serialize(m1)
        m2 = vistrails.core.db.io.unserialize(xml_str, Module)
        self.assertEquals(m1, m2)
        self.assertEquals(m1.id, m2.id)
        
    def testEq(self):
        """Check correctness of equality operator."""
        x = Module()
        self.assertNotEquals(x, None)

    def testAccessors(self):
        """Check that accessors are working."""
        x = Module()
        self.assertEquals(x.id, -1)
        x.id = 10
        self.assertEquals(x.id, 10)
        self.assertEquals(x.cache, 1)
        x.cache = 1
        self.assertEquals(x.cache, 1)
        self.assertEquals(x.location.x, -1.0)
        x.location = Location(x=1, y=x.location.y)
        self.assertEquals(x.location.x, 1)
        self.assertEquals(x.name, "")

    def testSummonModule(self):
        """Check that summon creates a correct module"""
        from vistrails.core.modules.basic_modules import identifier as basic_pkg

        x = Module()
        x.name = "String"
        x.package = basic_pkg
        try:
            registry = get_module_registry()
            c = x.summon()
            m = registry.get_descriptor_by_name(basic_pkg, 'String').module
            assert isinstance(c, m)
        except NoSummon:
            msg = "Expected to get a String object, got a NoSummon exception"
            self.fail(msg)

    def test_constructor(self):
        m1_param = ModuleParam(val="1.2",
                               type="Float",
                               alias="",
                               )
        m1_function = ModuleFunction(name="value",
                                     parameters=[m1_param],
                                     )
        m1 = Module(id=0,
                    name='Float',
                    functions=[m1_function],
                    )
                    
        m2 = Module()
        m2.name = "Float"
        m2.id = 0
        f = ModuleFunction()
        f.name = "value"
        m2.functions.append(f)
        param = ModuleParam()
        param.strValue = "1.2"
        param.type = "Float"
        param.alias = ""
        f.params.append(param)
        assert m1 == m2

    def test_str(self):
        m = Module(id=0,
                   name='Float',
                   functions=[ModuleFunction(name='value',
                                             parameters=[ModuleParam(type='Int',
                                                                     val='1',
                                                                     )],
                                             )],
                   )
        str(m)
        
if __name__ == '__main__':
    unittest.main()
    



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" File for the window used when opening VisTrails objects from the database

QOpenDBWindow
QDBConnectionList
QDBConnectionListItem
QDBObjectList
QDBObjectListItem
QConnectionDBSetupWindow

"""
from __future__ import division

from PyQt4 import QtCore, QtGui
from vistrails.db import VistrailsDBException
import vistrails.db.services.io
from vistrails.core.external_connection import ExtConnectionList, DBConnection
from vistrails.core.db.locator import DBLocator
from vistrails.core.system import default_connections_file
from vistrails.gui.theme import CurrentTheme
from vistrails.gui.utils import show_info, show_warning, show_question, \
NO_BUTTON, YES_BUTTON
from vistrails.core import debug

class QOpenDBWindow(QtGui.QDialog):
    """
    QOpenDBWindow is a dialog containing two panels. the left panel shows all
    the stored database connections and the right paanel shows the vistrails
    available on the selected database connection.

    """
    _instance = None
    def __init__(self, parent=None):
        """ __init__(parent: QWidget) -> QOpenDBWindow
        Construct the dialog with the two panels

        """
        QtGui.QDialog.__init__(self,parent)
        self.setWindowTitle("Choose a vistrail")
        self.save = False
        mainLayout = QtGui.QVBoxLayout()
        panelsLayout = QtGui.QGridLayout()

        self.createActions()
        self.saveasLayout = QtGui.QHBoxLayout()
        self.saveasLabel = QtGui.QLabel("Save As:")
        self.saveasEdt = QtGui.QLineEdit("")
        self.saveasEdt.setFixedWidth(200)
        self.saveasEdt.setSizePolicy(QtGui.QSizePolicy.Fixed,
                                     QtGui.QSizePolicy.Fixed)
        self.saveasLayout.addWidget(self.saveasLabel)
        self.saveasLayout.addWidget(self.saveasEdt)
        self.saveasLabel.setVisible(False)
        self.saveasEdt.setVisible(False)
        self.connectionList = QDBConnectionList(self)
        self.objectList = QDBObjectList(self)

        dbLabel = QtGui.QLabel("Databases:")
        self.vtLabel = QtGui.QLabel("Vistrails: ")
        
        panelsLayout.addWidget(dbLabel,0,0,1,1)
        panelsLayout.setColumnMinimumWidth(1,10)
        panelsLayout.addWidget(self.vtLabel,0,2,1,2)
        panelsLayout.addWidget(self.connectionList,1,0,1,1)
        panelsLayout.addWidget(self.objectList,1,2,1,2)

        self.addButton = QtGui.QToolButton()
        self.addButton.setDefaultAction(self.addAct)
        self.addButton.setAutoRaise(True)

        self.removeButton = QtGui.QToolButton()
        self.removeButton.setDefaultAction(self.removeAct)
        self.removeButton.setAutoRaise(True)
        self.removeButton.setEnabled(False)
        
        panelButtonsLayout = QtGui.QHBoxLayout()
        panelButtonsLayout.setMargin(0)
        panelButtonsLayout.setSpacing(0)
        panelButtonsLayout.addWidget(self.addButton)
        panelButtonsLayout.addWidget(self.removeButton)
        panelsLayout.addLayout(panelButtonsLayout,2,0,1,1,
                               QtCore.Qt.AlignLeft)
        buttonsLayout = QtGui.QHBoxLayout()
        self.cancelButton = QtGui.QPushButton('Cancel')
        self.cancelButton.setAutoDefault(False)
        self.openButton = QtGui.QPushButton('Open')
        self.openButton.setEnabled(False)
        self.openButton.setAutoDefault(True)
        
        buttonsLayout.addStretch(1)
        buttonsLayout.addWidget(self.cancelButton)
        buttonsLayout.addWidget(self.openButton)

        mainLayout.addLayout(self.saveasLayout)
        mainLayout.addLayout(panelsLayout)
        mainLayout.addLayout(buttonsLayout)
        mainLayout.setAlignment(self.saveasLayout, QtCore.Qt.AlignHCenter)
        self.setLayout(mainLayout)
        self.connectSignals()
        QOpenDBWindow._instance = self

    def createActions(self):
        """ createActions() -> None
        Create actions related to context menu 

        """
        self.addAct = QtGui.QAction("+", self)
        self.addAct.setStatusTip("Create a new connection")
        self.removeAct = QtGui.QAction("-", self)
        self.removeAct.setStatusTip("Remove the selected connection from list")
        
    def showEvent(self, e):
        """showEvent(e: QShowEvent) -> None
        If the use doesn't have any connection set up, we will ask him
        to create one.
        
        """
        if self.connectionList.count() == 0:
            text = "You don't seem to have any connection available. \
Would you like to create one?"
            res = show_question('Vistrails',
                                text, 
                                [NO_BUTTON, YES_BUTTON],
                                NO_BUTTON)
            if res == YES_BUTTON:
                self.showConnConfig()
        else:
            self.updateDBObjectsList()

    def connectSignals(self):
        """ connectSignals() -> None
        Map signals between GUI components        
        
        """
        self.connect(self.cancelButton,
                     QtCore.SIGNAL('clicked()'),
                     self.reject)
        self.connect(self.openButton,
                     QtCore.SIGNAL('clicked()'),
                     self.accept)
        self.connect(self.addAct,
                     QtCore.SIGNAL('triggered()'),
                     self.showConnConfig)
        self.connect(self.removeAct,
                     QtCore.SIGNAL('triggered()'),
                     self.connectionList.removeConnection)
        self.connect(self.connectionList,
                     QtCore.SIGNAL('itemSelectionChanged()'),
                     self.updateDBObjectsList)
        self.connect(self.connectionList,
                     QtCore.SIGNAL('itemSelectionChanged()'),
                     self.updateButtons)
        self.connect(self.connectionList,
                     QtCore.SIGNAL("reloadConnections"),
                     self.updateDBObjectsList)
        self.connect(self.objectList,
                     QtCore.SIGNAL('itemSelectionChanged()'),
                     self.updateButtons)
        self.connect(self.saveasEdt,
                     QtCore.SIGNAL('textChanged(QString)'),
                     self.updateButtons)
        self.connect(self.objectList,
                     QtCore.SIGNAL('itemDoubleClicked(QListWidgetItem *)'),
                     self.accept)

    def updateDBObjectsList(self):
        """ updateDBObjectsList() -> None
        It reloads the vistrails object list for the selected connection. 
        If nothing is selected, it will clear the list.

        """
        conn = self.connectionList.getCurrentItemId()
        try:
            self.objectList.updateContents(conn)
        except VistrailsDBException, e:
            self.connectionList.setCurrentItem(None)
        self.updateEditButtons(conn)

    def updateButtons(self):
        """updateButtons() -> None
        It will enable the open button if a vistrail is selected or in case
        of saving a vistrail, if a connection is selected and the name is valid

        """
        vtlist = self.objectList
        if not self.save:
            if len(vtlist.selectedItems()) > 0:
                self.openButton.setEnabled(True)
            else:
                self.openButton.setEnabled(False)
        else:
            if (len(self.connectionList.selectedItems()) > 0 and
                self.saveasEdt.text() != '' and
                len(vtlist.findItems(self.saveasEdt.text(),
                                     QtCore.Qt.MatchFixedString)) == 0):
                self.openButton.setEnabled(True)
            else:
                self.openButton.setEnabled(False)
                
    def updateEditButtons(self, id):
        """updateEditButtons(id: int) -> None
        It will enable/disable the connections buttons according to the
        selection

        """
        if id != -1:
            self.removeButton.setEnabled(True)
        else:
            self.removeButton.setEnabled(False)

    def showConnConfig(self, *args, **keywords):
        """showConnConfig(*args, **keywords) -> None
        shows a window to configure the connection. The valid keywords
        are defined in QConnectionDBSetupWindow.__init__()
        
        """
        keywords["parent"] = self
        
        dialog = QConnectionDBSetupWindow(**keywords)
        if dialog.exec_() == QtGui.QDialog.Accepted:
            config = {'id': int(dialog.id),
                      'name': str(dialog.nameEdt.text()),
                      'host': str(dialog.hostEdt.text()),
                      'port': int(dialog.portEdt.value()),
                      'user': str(dialog.userEdt.text()),
                      'passwd': str(dialog.passwdEdt.text()),
                      'db': str(dialog.databaseEdt.text())}
            id = self.connectionList.setConnectionInfo(**config)
            self.connectionList.setCurrentId(id)
            return True
        else:
            return False

    def prepareForOpening(self, obj_type):
        """prepareForOpening() -> None
        It will prepare the dialog to be a Open Dialog
        
        """
        self.setWindowTitle("Choose a %s" % obj_type.capitalize())
        if obj_type != 'vistrail':
            self.vtLabel.text = 'VisTrails %ss' % obj_type.capitalize()
        else:
            self.vtLabel.text = 'Vistrails'
        self.save = False
        self.objectList.obj_type = obj_type
        self.objectList.setEnabled(True)
        self.saveasLabel.setVisible(False)
        self.saveasEdt.setVisible(False)
        self.openButton.setEnabled(False)
        self.openButton.setText("Open")

    def prepareForSaving(self, obj_type):
        """prepareForSaving() -> None
        It will prepare the dialog to be a save as dialog 
        
        """
        self.setWindowTitle("Save %s..." % obj_type.capitalize())
        if obj_type != 'vistrail':
            self.vtLabel.text = 'VisTrails %ss' % obj_type.capitalize()
        else:
            self.vtLabel.text = 'Vistrails'
        self.save = True
        self.objectList.obj_type = obj_type
        self.objectList.setEnabled(False)
        self.saveasLabel.setVisible(True)
        self.saveasEdt.setVisible(True)
        self.openButton.setText("Save")
        self.openButton.setEnabled(False)
        
    @staticmethod
    def getInstance():
        """getInstance() -> QOpenDBWindow
        Returns the current instance of the QOpenDBWindow

        """
        if QOpenDBWindow._instance is None:
            QOpenDBWindow._instance = QOpenDBWindow()
        return QOpenDBWindow._instance
    
    @staticmethod
    def getOpenDBObject(obj_type):
        """getOpenDBObject(obj_type : str) -> (dict,int)
        Creates a dialog for opening a vistrails object from the database. 
        It will return the selected connection configuration information and
        the object id.
        
        """
        if QOpenDBWindow._instance:
            dlg = QOpenDBWindow._instance
        else:
            dlg = QOpenDBWindow()

        dlg.prepareForOpening(obj_type)
        
        if dlg.exec_() == QtGui.QDialog.Accepted:
            return (dlg.connectionList.getCurrentConnConfig(),
                    dlg.objectList.currentItem().id,
                    dlg.objectList.currentItem().name)
        else:
            return({},-1,'')

    @staticmethod
    def getSaveDBObject(obj_type):
        """getSaveDBObject(obj_type : str) -> (dict, str)
        Creates a dialog for saving a vistrails object to the database. 
        It will return the selected connection configuration information 
        and the object name

        """
        if QOpenDBWindow._instance:
            dlg = QOpenDBWindow._instance
        else:
            dlg = QOpenDBWindow()

        dlg.prepareForSaving(obj_type)

        if dlg.exec_() == QtGui.QDialog.Accepted:
            return (dlg.connectionList.getCurrentConnConfig(),
                    str(dlg.saveasEdt.text()).strip(' \n\t'))
        else:
            return({},'')
        
################################################################################

class QDBConnectionList(QtGui.QListWidget):
    """
    QDBConnection list is a widget to show the available databases

    """
    def __init__(self, parent=None):
        QtGui.QListWidget.__init__(self,parent)
        self.__list = ExtConnectionList.getInstance(default_connections_file())
        self.setSelectionMode(QtGui.QAbstractItemView.SingleSelection)
        self.setIconSize(QtCore.QSize(32,32))
        self.loadConnections()
        self.editAct = QtGui.QAction("Edit", self)
        self.editAct.setStatusTip("Edit the selected connection")
        self.connect(self.editAct,
                     QtCore.SIGNAL("triggered()"),
                     self.editConnection)
        
    def getCurrentItemId(self):
        """getCurrentItemId() -> int
        Returns the id of the selected item. If there is no item selected,
        it will return -1.

        """
        item = None
        if len(self.selectedItems()) > 0:
            item  = self.selectedItems()[0]
        if item is not None:
            return int(item.id)
        else:
            return -1
        
    def contextMenuEvent(self, e):
        """contextMenuEvent(e: QContextMenuEvent) -> None
        Shows a popup menu for the connection

        """
        item = self.currentItem()
        if item:
            menu = QtGui.QMenu()
            menu.addAction(self.editAct)
            menu.exec_(e.globalPos())

    def editConnection(self):
        """editConnection() -> None
        Method called to edit a connection. It will get the information
        from the selected connection and show the dialog so the user can
        update the fields
        
        """
        conn_id = self.getCurrentItemId()
        config = self.getConnectionInfo(conn_id)
        if config is not None:
            config["create"] = False
            self.parent().showConnConfig(**config)
            
    def updateGUI(self):
        """updateGUI() -> None
        Update GUI list to be consistent with the the list of connections

        """
        self.clear()
        for (id, c) in self.__list.items():
            cItem = QDBConnectionListItem(CurrentTheme.DB_ICON,
                                          int(id),
                                          str(c.name))
            self.addItem(cItem)
        self.emit(QtCore.SIGNAL("reloadConnections"))
        
    def loadConnections(self):
        """loadConnections() -> None
        Loads the internal connections and updates the GUI

        """
        self.__list.clear()
        self.__list.load_connections()
        self.updateGUI()

    def getConnectionInfo(self, id):
        """getConnectionInfo(id: int) -> dict
        Returns info of ExtConnection """
        conn = self.__list.get_connection(id)
        key = str(conn.id) + "." + conn.name + "." + conn.host
        passwd = DBLocator.keyChain.get_key(key)
        if conn is not None:
            config = {'id': conn.id,
                      'name': conn.name,
                      'host': conn.host,
                      'port': conn.port,
                      'user': conn.user,
                      'passwd': passwd,
                      'db': conn.database}
        else:
            config = None
        return config

    def findConnectionInfo(self, host, port, db):
        """findConnection(host:str, port: int, db: str) -> dict
        Returns complete info of a connection with the given parameters

        """
        id = self.__list.find_db_connection(host,port,db)
        if id != -1:
            return self.getConnectionInfo(id)
        else:
            return None
        
    def removeConnection(self):
        """removeConnection() -> None
        Removes the selected connection

        """
        id = self.getCurrentItemId()
        self.takeItem(self.currentRow())
        self.__list.remove_connection(id)
        
    def get_connection(self, id):
        """get_connection() -> None
        Returns the selected connection

        """
        return self.__list.get_connection(id)

    def setConnectionInfo(self, *args, **kwargs):
        """setConnectionInfo(id: int, name: str, host: str, port:int,
                     user:str, passwd:str, db:str) -> None
        If the connection exists it will update it, else it will add it

        """
        id = kwargs["id"]
        name = kwargs["name"]
        host = kwargs["host"]
        port = kwargs["port"]
        user = kwargs["user"]
        passwd = kwargs["passwd"]
        db = kwargs["db"]

        conn = DBConnection(id=id,
                            name=name,
                            host=host,
                            port=port,
                            user=user,
                            passwd='',
                            database=db,
                            dbtype='MySQL')
        
        if self.__list.has_connection(id):    
            self.__list.set_connection(id,conn)
        else:
            if conn.id == -1:
                conn.id = self.__list.get_fresh_id()
            self.__list.add_connection(conn)
        self.updateGUI()
        key = str(conn.id) + "." + conn.name + "." + conn.host
        DBLocator.keyChain.set_key(key,passwd)
        return conn.id
            
    def setCurrentId(self, id):
        """setCurrentId(id: int) -> None
        set the connection with id 'id' to be the current selected connection

        """
        conn = self.__list.get_connection(id)
        
        for i in self.findItems(conn.name, QtCore.Qt.MatchFixedString):
            if i.id == id:
                self.setCurrentItem(i)
                break
        self.emit(QtCore.SIGNAL("reloadConnections"), id)

    def getCurrentConnConfig(self):
        """getCurrentConnConfig() -> dict
        Return dictionary of parameters of the current connection to pass
        to MySQLdb

        """
        conn_id = self.currentItem().id
        conn = self.__list.get_connection(conn_id)
        config = self.getConnectionInfo(conn_id)
        if conn.dbtype == 'MySQL':
            #removing extra keyword arguments for MySQldb
            del config['name']
        return config
    
    def getDBObjectList(self, conn_id, obj_type):
        """getDBObjectList(conn_id: int, obj_type : str) -> list
        Returns list of vistrails objects

        """
        conn = self.__list.get_connection(conn_id)
        config = self.getConnectionInfo(conn_id)
        if conn.dbtype == 'MySQL':
            #removing extra keyword arguments for MySQldb
            config_name = config.pop('name')
            config_id = config.pop('id')
        vt_list = vistrails.db.services.io.get_db_object_list(config, obj_type)
        if conn.dbtype == 'MySQL':
            config['name'] = config_name
            config['id'] = config_id
        return vt_list
    
################################################################################
    
class QDBConnectionListItem(QtGui.QListWidgetItem):
    
    def __init__(self, icon, id, text, parent=None):
        """__init__(icon: QIcon, id: int, text: QString, parent: QListWidget)
                         -> QDBConnectionListItem
        Creates an item with id
        
        """
        QtGui.QListWidgetItem.__init__(self,icon, text, parent)
        self.id = id

################################################################################

class QDBObjectList(QtGui.QListWidget):
    """
    QDBObjectList is a widget to show the vistrails available in the selected
    database

    """
    def __init__(self, parent=None, obj_type='vistrail'):
        QtGui.QListWidget.__init__(self, parent)
        self.obj_type = obj_type

    def updateContents(self, conn_id=-1):
        """updateContents(connection_id: int) -> None
        Reloads vistrails from the given connection
        
        """
        self.clear()
        if conn_id != -1:
            parent = self.parent()
            try:
                objs = parent.connectionList.getDBObjectList(int(conn_id),
                                                             self.obj_type)
                
                for (id,obj,date) in objs:
                    item = QDBObjectListItem(CurrentTheme.FILE_ICON,
                                             int(id),
                                             str(obj),
                                             str(date))
                    self.addItem(item)
            except VistrailsDBException, e:
                #show connection setup
                if "Couldn't get list of vistrails objects" in str(e):
                    debug.critical('An error has occurred', e)
                    raise e
                config = parent.connectionList.getConnectionInfo(int(conn_id))
                if config is not None:
                    config["create"] = False
                    if not parent.showConnConfig(**config):
                        raise e
                else:
                    raise e
            
################################################################################

class QDBObjectListItem(QtGui.QListWidgetItem):
    
    def __init__(self, icon, id, name, date, parent=None):
        """__init__(icon: QIcon, id: int, name: QString,
                    date: QString, user: QString, parent: QListWidget)
                         -> QDBObjectListItem
        Creates an item with id
        
        """
        QtGui.QListWidgetItem.__init__(self, icon, name, parent)
        self.id = id
        self.name = name
        self.date = date
        self.setToolTip("Last Modified on %s" % date)

################################################################################

class QConnectionDBSetupWindow(QtGui.QDialog):
    """
    QConnectionDBSetupWindow is a dialog for creating a DB connection.
    
    """
    def __init__(self, parent=None, id=-1, name ='', host="", port=3306,
                 user="", passwd="", db="", create=True):
        """ __init__(parent: QWidget, id: int, name: str, host:str, port:int,
                     user:str, passwd:str, db:str, create:Boolean)
                                -> QConnectionDBSetupWindow
        Construct the dialog with the information provided
        create tells if the caption of the button is Create or Update

        """
        QtGui.QDialog.__init__(self,parent)
        if create:
            self.setWindowTitle("Create a new connection")
        else:
            self.setWindowTitle("Update a connection")
            
        mainLayout = QtGui.QVBoxLayout()
        infoLayout = QtGui.QGridLayout()
        self.id = id
        nameLabel = QtGui.QLabel("Save as Connection Name:", self)
        self.nameEdt = QtGui.QLineEdit(name, self)
        hostLabel = QtGui.QLabel("Server Hostname:", self)
        self.hostEdt = QtGui.QLineEdit(host, self)
        portLabel = QtGui.QLabel("Port:", self)
        self.portEdt = QtGui.QSpinBox(self)
        self.portEdt.setMaximum(65535)
        self.portEdt.setValue(port)
        userLabel = QtGui.QLabel("Username:", self)
        self.userEdt = QtGui.QLineEdit(user, self)
        passwdLabel = QtGui.QLabel("Password:", self)
        self.passwdEdt = QtGui.QLineEdit(passwd,self)
        self.passwdEdt.setEchoMode(QtGui.QLineEdit.Password)
        self.passwdEdt.setToolTip("For your protection, your "
                                  "password will not be saved.")
        databaseLabel = QtGui.QLabel("Database:", self)
        self.databaseEdt = QtGui.QLineEdit(db,self)
        mainLayout.addLayout(infoLayout)
        infoLayout.addWidget(nameLabel,0,0,1,1)
        infoLayout.addWidget(self.nameEdt,0,1,1,1)
        infoLayout.addWidget(hostLabel,1,0,1,1)
        infoLayout.addWidget(self.hostEdt,1,1,1,1)
        infoLayout.addWidget(portLabel,1,2,1,1)
        infoLayout.addWidget(self.portEdt,1,3,1,1)
        infoLayout.addWidget(userLabel,2,0,1,1)
        infoLayout.addWidget(self.userEdt,2,1,1,3)
        infoLayout.addWidget(passwdLabel,3,0,1,1)
        infoLayout.addWidget(self.passwdEdt,3,1,1,3)
        infoLayout.addWidget(databaseLabel,4,0,1,1)
        infoLayout.addWidget(self.databaseEdt,4,1,1,3)
        
        buttonsLayout = QtGui.QHBoxLayout()
        if create:
            caption = 'Create'
        else:
            caption = 'Update'
        self.createButton = QtGui.QPushButton(caption, self)
        self.createButton.setDefault(True)
        self.cancelButton = QtGui.QPushButton('Cancel', self)
        self.testButton = QtGui.QPushButton('Test', self)
        
        buttonsLayout.addStretch(1)
        buttonsLayout.addWidget(self.cancelButton)
        buttonsLayout.addWidget(self.testButton)
        buttonsLayout.addWidget(self.createButton)

        mainLayout.addLayout(buttonsLayout)

        self.setLayout(mainLayout)
        self.connectSignals()
        self.updateButtons()
        
    def connectSignals(self):
        """ connectSignals() -> None
        Map signals between GUI components        
        
        """
        self.connect(self.cancelButton,
                     QtCore.SIGNAL('clicked()'),
                     self.reject)
        self.connect(self.createButton,
                     QtCore.SIGNAL('clicked()'),
                     self.accept)
        self.connect(self.testButton,
                     QtCore.SIGNAL('clicked()'),
                     self.testConnection)
        self.connect(self.nameEdt,
                     QtCore.SIGNAL('textChanged(QString)'),
                     self.updateButtons)
        self.connect(self.hostEdt,
                     QtCore.SIGNAL('textChanged(QString)'),
                     self.updateButtons)
        self.connect(self.userEdt,
                     QtCore.SIGNAL('textChanged(QString)'),
                     self.updateButtons)
        self.connect(self.passwdEdt,
                     QtCore.SIGNAL('textChanged(QString)'),
                     self.updateButtons)
        self.connect(self.databaseEdt,
                     QtCore.SIGNAL('textChanged(QString)'),
                     self.updateButtons)
        self.connect(self.portEdt,
                     QtCore.SIGNAL('valueChanged(int)'),
                     self.updateButtons)

    def testConnection(self):
        """testConnection() -> None """
        config = {'host': str(self.hostEdt.text()),
                  'port': int(self.portEdt.value()),
                  'user': str(self.userEdt.text()),
                  'passwd': str(self.passwdEdt.text()),
                  'db': str(self.databaseEdt.text())}
        try:
            vistrails.db.services.io.test_db_connection(config)
            show_info('Vistrails',"Connection succeeded!")
            
        except Exception, e:
            debug.critical('An error has occurred', e)

    def updateButtons(self):
        """updateButtons() -> None
        enables button if there's enough information in the dialog

        """
        if (self.nameEdt.text() != "" and
            self.hostEdt.text() != "" and
            self.portEdt.value() != 0 and
            self.userEdt.text() != "" and
            self.databaseEdt.text() != ""):
            self.createButton.setEnabled(True)
        else:
            self.createButton.setEnabled(False)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

import os
from xml.dom import minidom, Node
from auto_gen_objects import Object, Property, Choice

class AutoGenParser(object):
    def __init__(self):
        pass

    def parse(self, dir):
        objects = {}
        for file in os.listdir(dir):
            if file.endswith('.xml'):
                filename = os.path.join(dir, file)
                # print filename
                dom = minidom.parse(filename)
                domObjects = dom.getElementsByTagName('object')
                for node in domObjects:
                    curObject = self.parseObject(node)
                    objects[curObject.getName()] = curObject

        # set the referenced objects here
        for obj in objects.itervalues():
            for prop in obj.properties:
                if prop.getReference():
                    try:
                        prop.setReferencedObject(objects[prop.getReference()])
                    except KeyError:
                        print 'error:', prop.getReference()
            for choice in obj.choices:
                for prop in choice.properties:
                    if prop.getReference():
                        try:
                            prop.setReferencedObject(
                                objects[prop.getReference()])
                        except KeyError:
                            print 'error:', prop.getReference()
                
        return objects.values()

    def parseObject(self, node):
        params = {}
        properties = []
        choices = []
        layouts = None
        for attr in node.attributes.keys():
            params[attr] = node.attributes.get(attr).value
        for child in node.childNodes:
            if child.nodeType == Node.ELEMENT_NODE:
                if child.nodeName == 'layout':
                    layouts = self.parseLayouts(child)
                elif child.nodeName == 'property':
                    property = self.parseProperty(child)
                    properties.append(property)
                elif child.nodeName == 'choice':
                    choice = self.parseChoice(child)
                    choices.append(choice)
        return Object(params, properties, layouts, choices)
    
    def parseLayouts(self, node):
        layouts = {}
        for child in node.childNodes:
            if child.nodeType == Node.ELEMENT_NODE:
                layouts[child.nodeName] = self.parseDataToDict(child)
        return layouts

    def parseProperty(self, node):
        params = {}
        specs = {}
        for attr in node.attributes.keys():
            params[attr] = node.attributes.get(attr).value
        for child in node.childNodes:
            if child.nodeType == Node.ELEMENT_NODE:
                specs[child.nodeName] = self.parseDataToDict(child)
        return Property(params, specs)

    def parseChoice(self, node):
        params = {}
        properties = []
        for attr in node.attributes.keys():
            params[attr] = node.attributes.get(attr).value
        for child in node.childNodes:
            if child.nodeType == Node.ELEMENT_NODE and \
                    child.nodeName == 'property':
                properties.append(self.parseProperty(child))
        return Choice(params, properties)

    def parseDataToDict(self, node):
        dict = {}
        if node.nodeType == Node.ELEMENT_NODE:
            for attr in node.attributes.keys():
                dict[attr] = node.attributes.get(attr).value
            for child in node.childNodes:
                if child.nodeType == Node.ELEMENT_NODE:
                    dict[child.nodeName] = child.childNodes[0].data
        return dict



###############################################################################

##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This file contains a dialog for editing options for how the given
    VisTrails module is executed.

QModuleOptions
"""
from __future__ import division

from PyQt4 import QtCore, QtGui
from vistrails.core.vistrail.module_control_param import ModuleControlParam
from vistrails.gui.theme import CurrentTheme
from vistrails.gui.vistrails_palette import QVistrailsPaletteInterface

import json
import unittest

###############################################################################

class QModuleOptions(QtGui.QDialog, QVistrailsPaletteInterface):
    """
    QModuleIteration is a dialog for editing module looping options.

    """
    def __init__(self, parent=None):
        """ 
        QModuleIteration(parent)
        -> None

        """
        QtGui.QDialog.__init__(self, parent)
        self.setWindowTitle("Module Execution Options")
        self.createButtons()
        self.update_module()

    def createButtons(self):
        """ createButtons() -> None
        Create and connect signals to Ok & Cancel button
        
        """
        self.controller = None
        self.state_changed = False
        self.module = None
        self.setLayout(QtGui.QVBoxLayout())
        # self.layout().addStrut()
        layout = QtGui.QHBoxLayout()
        type_group = QtGui.QButtonGroup(self) # Number group
        layout.addWidget(QtGui.QLabel("Port list combination method:"))
        self.pairwiseButton = QtGui.QRadioButton("Pairwise")
        self.pairwiseButton.setToolTip("Execute multiple looped input ports pairwise:"
                                       " [(A, B), (C, D)] -> [(A, C), (B, D)]")
        type_group.addButton(self.pairwiseButton)
        layout.addWidget(self.pairwiseButton)
        layout.setStretch(0, 0)
        self.cartesianButton = QtGui.QRadioButton("Cartesian")
        self.cartesianButton.setToolTip("Execute multiple looped input ports using cartesian product:"
                                       " [(A, B), (C, D)] -> [(A, C), (A, D), (B, C), (B, D)]")
        self.cartesianButton.setChecked(True)
        type_group.addButton(self.cartesianButton)
        layout.addWidget(self.cartesianButton)
        layout.setStretch(1, 0)
        self.customButton = QtGui.QRadioButton("Custom")
        self.customButton.setToolTip("Build a custom combination using pairwise/cartesian functions")
        type_group.addButton(self.customButton)
        layout.addWidget(self.customButton)
        layout.setStretch(2, 0)
        layout.addStretch(1)
        self.layout().addLayout(layout)
        self.layout().setStretch(0, 0)

        self.portCombiner = QPortCombineTreeWidget()
        self.layout().addWidget(self.portCombiner)
        self.portCombiner.setVisible(False)
        
        whileLayout = QtGui.QVBoxLayout()

        self.whileButton = QtGui.QCheckBox("While Loop")
        self.whileButton.setToolTip('Repeatedly execute module until a specified output port has a false value')
        whileLayout.addWidget(self.whileButton)
        whileLayout.setStretch(0, 0)

        layout = QtGui.QHBoxLayout()
        self.condLabel = QtGui.QLabel("Condition output port:")
        layout.addWidget(self.condLabel)
        layout.setStretch(0, 0)
        self.condEdit = QtGui.QLineEdit()
        self.condEdit.setToolTip('Name of output port containing the condition of the loop')
        layout.addWidget(self.condEdit)
        layout.setStretch(1, 1)
        whileLayout.addLayout(layout)
        whileLayout.setStretch(1, 0)

        layout = QtGui.QHBoxLayout()
        self.maxLabel = QtGui.QLabel("Max iterations:")
        layout.addWidget(self.maxLabel)
        layout.setStretch(0, 0)
        self.maxEdit = QtGui.QLineEdit()
        self.maxEdit.setValidator(QtGui.QIntValidator())
        self.maxEdit.setToolTip('Fail after this number of iterations have been reached (default=20)')
        layout.addWidget(self.maxEdit)
        layout.setStretch(1, 1)
        whileLayout.addLayout(layout)
        whileLayout.setStretch(2, 0)

        layout = QtGui.QHBoxLayout()
        self.delayLabel = QtGui.QLabel("Delay:")
        layout.addWidget(self.delayLabel)
        layout.setStretch(0, 0)
        self.delayEdit = QtGui.QLineEdit()
        self.delayEdit.setValidator(QtGui.QDoubleValidator(self))
        self.delayEdit.setToolTip('Delay between iterations in fractions of seconds')
        layout.addWidget(self.delayEdit)
        layout.setStretch(1, 1)
        whileLayout.addLayout(layout)
        whileLayout.setStretch(2, 0)

        layout = QtGui.QHBoxLayout()
        self.feedInputLabel = QtGui.QLabel("Feedback Input port:")
        layout.addWidget(self.feedInputLabel)
        layout.setStretch(0, 0)
        self.feedInputEdit = QtGui.QLineEdit()
        self.feedInputEdit.setToolTip('Name of input port to feed the value from last iteration')
        layout.addWidget(self.feedInputEdit)
        layout.setStretch(1, 1)
        whileLayout.addLayout(layout)
        whileLayout.setStretch(3, 0)

        layout = QtGui.QHBoxLayout()
        self.feedOutputLabel = QtGui.QLabel("Feedback Output port:")
        layout.addWidget(self.feedOutputLabel)
        layout.setStretch(0, 0)
        self.feedOutputEdit = QtGui.QLineEdit()
        self.feedOutputEdit.setToolTip('Name of output port to feed to next iteration')
        layout.addWidget(self.feedOutputEdit)
        layout.setStretch(1, 1)
        whileLayout.addLayout(layout)
        whileLayout.setStretch(4, 0)

        whileLayout.addStretch(1)
        self.layout().addLayout(whileLayout)

        self.jobCacheButton = QtGui.QCheckBox("Cache Output Persistently")
        self.jobCacheButton.setToolTip('Cache the module results persistently to disk. (outputs must be constants)')
        self.layout().addWidget(self.jobCacheButton)
        self.layout().setStretch(2, 0)

        self.layout().addStretch(1)
        self.buttonLayout = QtGui.QHBoxLayout()
        self.buttonLayout.setMargin(5)
        self.saveButton = QtGui.QPushButton('&Save', self)
        self.saveButton.setFixedWidth(100)
        self.saveButton.setEnabled(False)
        self.buttonLayout.addWidget(self.saveButton)
        self.resetButton = QtGui.QPushButton('&Reset', self)
        self.resetButton.setFixedWidth(100)
        self.resetButton.setEnabled(False)
        self.buttonLayout.addWidget(self.resetButton)
        self.layout().addLayout(self.buttonLayout)
        self.connect(self.saveButton, QtCore.SIGNAL('clicked(bool)'),
                     self.saveTriggered)
        self.connect(self.resetButton, QtCore.SIGNAL('clicked(bool)'),
                     self.resetTriggered)        
        self.layout().setStretch(3, 0)
        self.update_module()
        self.pairwiseButton.toggled.connect(self.stateChanged)
        self.cartesianButton.toggled.connect(self.stateChanged)
        self.customButton.toggled.connect(self.stateChanged)
        self.customButton.toggled.connect(self.customToggled)
        self.portCombiner.itemChanged.connect(self.stateChanged)
        self.whileButton.toggled.connect(self.stateChanged)
        self.whileButton.toggled.connect(self.whileToggled)
        self.condEdit.textChanged.connect(self.stateChanged)
        self.maxEdit.textChanged.connect(self.stateChanged)
        self.delayEdit.textChanged.connect(self.stateChanged)
        self.feedInputEdit.textChanged.connect(self.stateChanged)
        self.feedOutputEdit.textChanged.connect(self.stateChanged)
        self.jobCacheButton.toggled.connect(self.stateChanged)

    def sizeHint(self):
        """ sizeHint() -> QSize
        Return the recommended size of the configuration window
        
        """
        return QtCore.QSize(512, 256)

    def saveTriggered(self, checked = False):
        """ saveTriggered(checked: bool) -> None
        Update vistrail controller and module when the user click Ok
        
        """
        if self.updateVistrail():
            self.saveButton.setEnabled(False)
            self.resetButton.setEnabled(False)
            self.state_changed = False
            self.emit(QtCore.SIGNAL("stateChanged"))
            self.emit(QtCore.SIGNAL('doneConfigure'), self.module.id)
            
    def resetTriggered(self, checked = False):
        self.update_module(self.module)

    def stateChanged(self, state=False, other=None):
        self.saveButton.setEnabled(True)
        self.resetButton.setEnabled(True)
        self.state_changed = True

    def customToggled(self, state=False):
        self.portCombiner.setVisible(state)

    def whileToggled(self, state=False):
        if state:
            self.condEdit.setVisible(True)
            self.maxEdit.setVisible(True)
            self.delayEdit.setVisible(True)
            self.feedInputEdit.setVisible(True)
            self.feedOutputEdit.setVisible(True)
            self.condLabel.setVisible(True)
            self.maxLabel.setVisible(True)
            self.delayLabel.setVisible(True)
            self.feedInputLabel.setVisible(True)
            self.feedOutputLabel.setVisible(True)
            self.condEdit.setText('')
            self.maxEdit.setText('')
            self.delayEdit.setText('')
            self.feedInputEdit.setText('')
            self.feedOutputEdit.setText('')
        else:
            self.condEdit.setVisible(False)
            self.maxEdit.setVisible(False)
            self.delayEdit.setVisible(False)
            self.feedInputEdit.setVisible(False)
            self.feedOutputEdit.setVisible(False)
            self.condLabel.setVisible(False)
            self.maxLabel.setVisible(False)
            self.delayLabel.setVisible(False)
            self.feedInputLabel.setVisible(False)
            self.feedOutputLabel.setVisible(False)
    
    def closeEvent(self, event):
        self.askToSaveChanges()
        event.accept()

    def set_controller(self, controller):
        self.controller = controller
        if not controller:
            return
        scene = controller.current_pipeline_scene
        selected_ids = scene.get_selected_module_ids() 
        modules = [controller.current_pipeline.modules[i] 
                   for i in selected_ids]
        if len(modules) == 1:
            self.update_module(modules[0])
        else:
            self.update_module(None)

    def update_module(self, module=None):
        self.module = module
        if not module:
            self.pairwiseButton.setEnabled(False)
            self.cartesianButton.setEnabled(False)
            self.customButton.setEnabled(False)
            self.whileButton.setEnabled(False)
            self.condEdit.setVisible(False)
            self.maxEdit.setVisible(False)
            self.delayEdit.setVisible(False)
            self.feedInputEdit.setVisible(False)
            self.feedOutputEdit.setVisible(False)
            self.condLabel.setVisible(False)
            self.maxLabel.setVisible(False)
            self.delayLabel.setVisible(False)
            self.feedInputLabel.setVisible(False)
            self.feedOutputLabel.setVisible(False)
            self.portCombiner.setVisible(False)
            self.jobCacheButton.setEnabled(False)
            self.state_changed = False
            self.saveButton.setEnabled(False)
            self.resetButton.setEnabled(False)
            return
        # set defaults
        self.pairwiseButton.setEnabled(True)
        self.cartesianButton.setEnabled(True)
        self.cartesianButton.setChecked(True)
        self.customButton.setEnabled(True)

        self.whileButton.setEnabled(True)
        self.whileButton.setChecked(False)
        self.condEdit.setVisible(False)
        self.maxEdit.setVisible(False)
        self.delayEdit.setVisible(False)
        self.feedInputEdit.setVisible(False)
        self.feedOutputEdit.setVisible(False)
        self.condLabel.setVisible(False)
        self.maxLabel.setVisible(False)
        self.delayLabel.setVisible(False)
        self.feedInputLabel.setVisible(False)
        self.feedOutputLabel.setVisible(False)
        self.portCombiner.setVisible(False)
        self.portCombiner.setDefault(module)
        self.jobCacheButton.setEnabled(True)
        self.jobCacheButton.setChecked(False)
        if module.has_control_parameter_with_name(ModuleControlParam.LOOP_KEY):
            type = module.get_control_parameter_by_name(ModuleControlParam.LOOP_KEY).value
            self.pairwiseButton.setChecked(type=='pairwise')
            self.cartesianButton.setChecked(type=='cartesian')
            self.customButton.setChecked(type not in ['pairwise', 'cartesian'])
            self.portCombiner.setVisible(type not in ['pairwise', 'cartesian'])
            if type not in ['pairwise', 'cartesian']:
                self.portCombiner.setValue(type)
        if module.has_control_parameter_with_name(ModuleControlParam.WHILE_COND_KEY) or \
           module.has_control_parameter_with_name(ModuleControlParam.WHILE_MAX_KEY):
            self.whileButton.setChecked(True)
        if module.has_control_parameter_with_name(ModuleControlParam.WHILE_COND_KEY):
            cond = module.get_control_parameter_by_name(ModuleControlParam.WHILE_COND_KEY).value
            self.condEdit.setText(cond)
        if module.has_control_parameter_with_name(ModuleControlParam.WHILE_MAX_KEY):
            max = module.get_control_parameter_by_name(ModuleControlParam.WHILE_MAX_KEY).value
            self.maxEdit.setText(max)
        if module.has_control_parameter_with_name(ModuleControlParam.WHILE_DELAY_KEY):
            delay = module.get_control_parameter_by_name(ModuleControlParam.WHILE_DELAY_KEY).value
            self.delayEdit.setText(delay)
        if module.has_control_parameter_with_name(ModuleControlParam.WHILE_INPUT_KEY):
            input = module.get_control_parameter_by_name(ModuleControlParam.WHILE_INPUT_KEY).value
            self.feedInputEdit.setText(input)
        if module.has_control_parameter_with_name(ModuleControlParam.WHILE_OUTPUT_KEY):
            output = module.get_control_parameter_by_name(ModuleControlParam.WHILE_OUTPUT_KEY).value
            self.feedOutputEdit.setText(output)
        if module.has_control_parameter_with_name(ModuleControlParam.JOB_CACHE_KEY):
            jobCache = module.get_control_parameter_by_name(ModuleControlParam.JOB_CACHE_KEY).value
            self.jobCacheButton.setChecked(jobCache.lower()=='true')
        self.state_changed = False
        self.saveButton.setEnabled(False)
        self.resetButton.setEnabled(False)

    def updateVistrail(self):
        values = []
        if self.pairwiseButton.isChecked():
            value = 'pairwise'
        elif self.cartesianButton.isChecked():
            value = 'cartesian'
        else:
            value = self.portCombiner.getValue()
        values.append((ModuleControlParam.LOOP_KEY, value))
        _while = self.whileButton.isChecked()
        values.append((ModuleControlParam.WHILE_COND_KEY,
                       _while and self.condEdit.text()))
        values.append((ModuleControlParam.WHILE_MAX_KEY,
                       _while and self.maxEdit.text()))
        values.append((ModuleControlParam.WHILE_DELAY_KEY,
                       _while and self.delayEdit.text()))
        values.append((ModuleControlParam.WHILE_INPUT_KEY,
                       _while and self.feedInputEdit.text()))
        values.append((ModuleControlParam.WHILE_OUTPUT_KEY,
                       _while and self.feedOutputEdit.text()))
        jobCache = self.jobCacheButton.isChecked()
        values.append((ModuleControlParam.JOB_CACHE_KEY,
                       [False, 'true'][jobCache]))
        for name, value in values:
            if value:
                if (not self.module.has_control_parameter_with_name(name) or
                        value != self.module.get_control_parameter_by_name(name).value):
                    if self.module.has_control_parameter_with_name(name):
                        self.controller.delete_control_parameter(name,
                                                                 self.module.id)
                    self.controller.add_control_parameter((name, value),
                                                          self.module.id)
            elif self.module.has_control_parameter_with_name(name):
                self.controller.delete_control_parameter(name, self.module.id)
        return True

    def activate(self):
        if self.isVisible() == False:
            self.show()
        self.activateWindow()

PORTITEM = 1000
DOTITEM = 1001
CROSSITEM = 1002
class PortItem(QtGui.QTreeWidgetItem):
    def __init__(self, port_name, parent=None):
        QtGui.QTreeWidgetItem.__init__(self, parent, PORTITEM)
        self.setText(0, port_name)
        self.setFlags(self.flags() & ~QtCore.Qt.ItemIsDropEnabled)

class DotItem(QtGui.QTreeWidgetItem):
    def __init__(self, parent=None):
        QtGui.QTreeWidgetItem.__init__(self, parent, DOTITEM)
        self.setExpanded(True)
        self.setIcon(0, CurrentTheme.DOT_PRODUCT_ICON)
        self.setText(0, 'Dot')

class CrossItem(QtGui.QTreeWidgetItem):
    def __init__(self, parent=None):
        QtGui.QTreeWidgetItem.__init__(self, parent, CROSSITEM)
        self.setExpanded(True)
        self.setIcon(0, CurrentTheme.CROSS_PRODUCT_ICON)
        self.setText(0, 'Cross')

class QPortCombineTreeWidget(QtGui.QTreeWidget):
    def __init__(self, parent=None):
        QtGui.QTreeWidget.__init__(self, parent)
        self.setDragDropMode(QtGui.QAbstractItemView.InternalMove)
        self.header().hide()
        self.setExpandsOnDoubleClick(False)
        self.setItemsExpandable(False)
        self.setRootIsDecorated(False)
        self.expandAll()
        self.setToolTip("Right-click to add dot/cross product. Rearrange "
                        "items to get suitable order. 'Del' key deletes "
                        "selected product.")

    def dropEvent(self, event):
        QtGui.QTreeWidget.dropEvent(self, event)
        self.expandAll()

    def loadNode(self, parent, node):
        # populate widget from json struct
        if isinstance(node, basestring):
            PortItem(node, parent)
        else:
            item = DotItem(parent) if node[0] == 'pairwise' \
                                   else CrossItem(parent)
            for i in node[1:]:
                self.loadNode(item, i)
        
    def saveNode(self, item):
        # populate json struct from widget items
        if item.type()==PORTITEM:
            return item.text(0)
        L = ['pairwise'] if item.type()==DOTITEM else ['cartesian']
        L.extend([self.saveNode(item.child(i))
                  for i in xrange(item.childCount())])
        L = [i for i in L if i is not None]
        if len(L)<2:
            L = None
        return L
    
    def setValue(self, value):
        self.clear()
        value = json.loads(value)
        for v in value[1:]:
            self.loadNode(self.invisibleRootItem(), v)
    
    def getValue(self):
        nodes = [self.topLevelItem(i)
                 for i in xrange(self.topLevelItemCount())]
        L = ['cartesian'] # default
        L.extend([self.saveNode(node) for node in nodes])
        L = [i for i in L if i is not None]
        if len(L)<2:
            L = None
        return json.dumps(L)

    def setDefault(self, module):
        self.clear()
        if not module:
            return
        for port_name in module.iterated_ports:
            PortItem(port_name, self)
            
    def contextMenuEvent(self, event):
        menu = QtGui.QMenu()
        dotAction = QtGui.QAction(CurrentTheme.DOT_PRODUCT_ICON,
                                  'Add Pairwise Product', self)
        dotAction.triggered.connect(self.addDot)
        menu.addAction(dotAction)
        crossAction = QtGui.QAction(CurrentTheme.CROSS_PRODUCT_ICON,
                                    'Add Cartesian Product', self)
        crossAction.triggered.connect(self.addCross)
        menu.addAction(crossAction)
        menu.exec_(event.globalPos())
        event.accept()

    def addDot(self):
        DotItem(self)

    def addCross(self):
        CrossItem(self)

    def keyPressEvent(self, event):
        """ keyPressEvent(event: QKeyEvent) -> None
        Capture 'Del', 'Backspace' for deleting items.
        Ctrl+C, Ctrl+V, Ctrl+A for copy, paste and select all
        
        """
        items = self.selectedItems()
        if (len(items)==1 and \
            event.key() in [QtCore.Qt.Key_Backspace, QtCore.Qt.Key_Delete]) and \
            type(items[0]) in [DotItem, CrossItem] and\
            not items[0].childCount():
            item = items[0]
            if item.parent():
                item.parent().takeChild(item.parent().indexOfChild(item))
            else:
                self.takeTopLevelItem(self.indexOfTopLevelItem(item))
        else:
            QtGui.QTreeWidget.keyPressEvent(self, event)

class TestIterationGui(unittest.TestCase):
    def testGetSet(self):
        p = QPortCombineTreeWidget()
        v = '["cartesian", ["pairwise", "a", "b"], "c"]'
        p.setValue(v)
        self.assertEqual(v, p.getValue())
